from klvm import KEYWORD_TO_ATOM

from klvm_tools.pattern_match import match
from klvm_tools.binutils import assemble

from klvm_tools.NodePath import NodePath, LEFT, RIGHT
from .helpers import quote

QUOTE_ATOM = KEYWORD_TO_ATOM["q"]
APPLY_ATOM = KEYWORD_TO_ATOM["a"]
FIRST_ATOM = KEYWORD_TO_ATOM["f"]
REST_ATOM = KEYWORD_TO_ATOM["r"]
CONS_ATOM = KEYWORD_TO_ATOM["c"]
RAISE_ATOM = KEYWORD_TO_ATOM["x"]

DEBUG_OPTIMIZATIONS = 0


def non_nil(sexp):
    return sexp.listp() or len(sexp.as_atom()) > 0


def seems_constant(sexp):
    if not sexp.listp():
        # note that `0` is a constant
        return not non_nil(sexp)
    operator = sexp.first()
    if not operator.listp():
        as_atom = operator.as_atom()
        if as_atom == QUOTE_ATOM:
            return True
        if as_atom == RAISE_ATOM:
            return False
    elif not seems_constant(operator):
        return False
    return all(seems_constant(_) for _ in sexp.rest().as_iter())


def constant_optimizer(r, eval):
    """
    If the expression does not depend upon @ anywhere,
    it's a constant. So we can simply evaluate it and
    return the quoted result.
    """
    if seems_constant(r) and non_nil(r):
        cost, r1 = eval(r, r.null())
        r = r.to(quote(r1))
    return r


def is_args_call(r):
    if not r.listp() and r.as_int() == 1:
        return True
    return False


CONS_Q_A_OPTIMIZER_PATTERN = assemble("(a (q . (: . sexp)) (: . args))")


def cons_q_a_optimizer(r, eval):
    """
    This applies the transform
    (a (q . SEXP) @) => SEXP
    """
    t1 = match(CONS_Q_A_OPTIMIZER_PATTERN, r)
    if t1 and is_args_call(t1["args"]):
        return t1["sexp"]
    return r


CONS_PATTERN = assemble("(c (: . first) (: . rest)))")


def cons_f(args):
    t = match(CONS_PATTERN, args)
    if t:
        return t["first"]
    return args.to([FIRST_ATOM, args])


def cons_r(args):
    t = match(CONS_PATTERN, args)
    if t:
        return t["rest"]
    return args.to([REST_ATOM, args])


def path_from_args(sexp, new_args):
    v = sexp.as_int()
    if v <= 1:
        return new_args
    sexp = sexp.to(v >> 1)
    if v & 1:
        return path_from_args(sexp, cons_r(new_args))
    return path_from_args(sexp, cons_f(new_args))


def sub_args(sexp, new_args):
    if sexp.nullp() or not sexp.listp():
        return path_from_args(sexp, new_args)

    first = sexp.first()
    if first.listp():
        first = sub_args(first, new_args)
    else:
        op = first.as_atom()

        if op == QUOTE_ATOM:
            return sexp

    return sexp.to([first] + [sub_args(_, new_args) for _ in sexp.rest().as_iter()])


VAR_CHANGE_OPTIMIZER_CONS_EVAL_PATTERN = assemble("(a (q . (: . sexp)) (: . args))")


def var_change_optimizer_cons_eval(r, eval):
    """
    This applies the transform
    (a (q . (op SEXP1...)) (ARGS)) => (q . RET_VAL) where ARGS != @
    via
    (op (a SEXP1 (ARGS)) ...) (ARGS)) and then "children_optimizer" of this.
    In some cases, this can result in a constant in some of the children.

    If we end up needing to push the "change of variables" to only one child, keep
    the optimization. Otherwise discard it.
    """

    t1 = match(VAR_CHANGE_OPTIMIZER_CONS_EVAL_PATTERN, r)

    if t1 is None:
        return r

    original_args = t1["args"]

    original_call = t1["sexp"]

    new_eval_sexp_args = sub_args(original_call, original_args)

    # Do not iterate into a quoted value as if it were a list
    if seems_constant(new_eval_sexp_args):
        new_operands = new_eval_sexp_args
        opt_operands = optimize_sexp(new_operands, eval)
        return r.to(opt_operands)

    new_operands = list(new_eval_sexp_args.as_iter())
    opt_operands = [optimize_sexp(_, eval) for _ in new_operands]
    non_constant_count = sum(1 if _.listp() and _.first() != QUOTE_ATOM else 0 for _ in opt_operands)
    if non_constant_count < 1:
        final_sexp = r.to(opt_operands)
        return final_sexp
    return r


def children_optimizer(r, eval):
    """
    Recursively apply optimizations to all non-quoted child nodes.
    """
    if not r.listp():
        return r
    operator = r.first()
    if not operator.listp():
        op = operator.as_atom()
        if op == QUOTE_ATOM:
            return r
    return r.to([optimize_sexp(_, eval) for _ in r.as_iter()])


CONS_OPTIMIZER_PATTERN_FIRST = assemble("(f (c (: . first) (: . rest)))")
CONS_OPTIMIZER_PATTERN_REST = assemble("(r (c (: . first) (: . rest)))")


def cons_optimizer(r, eval):
    """
    This applies the transform
    (f (c A B)) => A
    and
    (r (c A B)) => B
    """
    t1 = match(CONS_OPTIMIZER_PATTERN_FIRST, r)
    if t1:
        return t1["first"]
    t1 = match(CONS_OPTIMIZER_PATTERN_REST, r)
    if t1:
        return t1["rest"]
    return r


FIRST_ATOM_PATTERN = assemble("(f ($ . atom))")
REST_ATOM_PATTERN = assemble("(r ($ . atom))")


def path_optimizer(r, eval):
    """
    This applies the transform
    (f N) => A
    and
    (r N) => B
    """

    t1 = match(FIRST_ATOM_PATTERN, r)
    if t1 and non_nil(t1["atom"]):
        node = NodePath(t1["atom"].as_int())
        node = node + LEFT
        return r.to(node.as_short_path())

    t1 = match(REST_ATOM_PATTERN, r)
    if t1 and non_nil(t1["atom"]):
        node = NodePath(t1["atom"].as_int())
        node = node + RIGHT
        return r.to(node.as_short_path())
    return r


QUOTE_PATTERN_1 = assemble("(q . 0)")


def quote_null_optimizer(r, eval):
    """
    This applies the transform `(q . 0)` => `0`
    """
    t1 = match(QUOTE_PATTERN_1, r)
    if t1 is not None:
        return r.to(0)

    return r


APPLY_NULL_PATTERN_1 = assemble("(a 0 . (: . rest))")


def apply_null_optimizer(r, eval):
    """
    This applies the transform `(a 0 ARGS)` => `0`
    """
    t1 = match(APPLY_NULL_PATTERN_1, r)
    if t1 is not None:
        return r.to(0)

    return r


def optimize_sexp(r, eval):
    """
    Optimize an s-expression R written for klvm to R_opt where
    (a R args) == (a R_opt args) for ANY args.
    """
    if r.nullp() or not r.listp():
        return r

    OPTIMIZERS = [
        cons_optimizer,
        constant_optimizer,
        cons_q_a_optimizer,
        var_change_optimizer_cons_eval,
        children_optimizer,
        path_optimizer,
        quote_null_optimizer,
        apply_null_optimizer,
    ]

    while r.listp():
        start_r = r
        for opt in OPTIMIZERS:
            r = opt(r, eval)
            if start_r != r:
                break
        if start_r == r:
            return r
        if DEBUG_OPTIMIZATIONS:
            print("OPT-%s[%s] => %s\n" % (
                opt.__name__, start_r, r))
    return r


def make_do_opt(run_program):

    def do_opt(args):
        return 1, optimize_sexp(args.first(), run_program)

    return do_opt



from klvm import KEYWORD_TO_ATOM

from klvm_tools import binutils
from klvm_tools.debug import build_symbol_dump
from klvm_tools.NodePath import LEFT, RIGHT, TOP

from .helpers import eval, quote
from .optimize import optimize_sexp


QUOTE_ATOM = KEYWORD_TO_ATOM["q"]
CONS_ATOM = KEYWORD_TO_ATOM["c"]

MAIN_NAME = b""


def build_tree(items):
    """
    This function takes a Python list of items and turns it into a binary tree
    of the items, suitable for casting to an s-expression.
    """
    size = len(items)
    if size == 0:
        return []
    if size == 1:
        return items[0]
    half_size = size >> 1
    left = build_tree(items[:half_size])
    right = build_tree(items[half_size:])
    return (left, right)


def build_tree_program(items):
    """
    This function takes a Python list of items and turns it into a program that
    builds a binary tree of the items, suitable for casting to an s-expression.
    """
    size = len(items)
    if size == 0:
        return [quote([])]
    if size == 1:
        return items[0]
    half_size = size >> 1
    left = build_tree_program(items[:half_size])
    right = build_tree_program(items[half_size:])
    return [CONS_ATOM, left, right]


def flatten(sexp):
    """
    Return a (python) list of every atom.
    """
    if sexp.listp():
        r = []
        r.extend(flatten(sexp.first()))
        r.extend(flatten(sexp.rest()))
        return r
    return [sexp.as_atom()]


def build_used_constants_names(functions, constants, macros):
    """
    Do a naïve pruning of unused symbols. It may be too big, but it shouldn't
    be too small. Return a list of all atoms used that are also the names of
    functions or constants, starting with the MAIN_NAME function.
    """
    macros_as_dict = {_.rest().first().as_atom(): _ for _ in macros}

    possible_symbols = set(functions.keys())
    possible_symbols.update(constants.keys())

    new_names = set([MAIN_NAME])
    used_names = set(new_names)
    while new_names:
        prior_new_names = new_names
        new_names = set()
        for _ in prior_new_names:
            for k in [functions, macros_as_dict]:
                if _ in k:
                    new_names.update(flatten(k[_]))
        new_names.difference_update(used_names)
        used_names.update(new_names)
    used_names.intersection_update(possible_symbols)
    used_names.discard(MAIN_NAME)
    return sorted(used_names)


def parse_include(name, namespace, functions, constants, macros, run_program):
    prog = binutils.assemble("(_read (_full_path_for_name 1))")
    cost, assembled_sexp = run_program(prog, name)
    for sexp in assembled_sexp.as_iter():
        parse_mod_sexp(sexp, namespace, functions, constants, macros, run_program)


def unquote_args(code, args):
    if code.listp():
        c1 = code.first()
        c2 = code.rest()
        return unquote_args(c1, args).cons(unquote_args(c2, args))

    if code.as_atom() in args:
        return code.to([b"unquote", code])

    return code


def defun_inline_to_macro(declaration_sexp):
    d2 = declaration_sexp.rest()
    d3 = d2.rest()
    r = [b"defmacro", d2.first(), d3.first()]
    code = d3.rest().first()
    args = [_ for _ in flatten(d3.first()) if _ != b""]
    unquoted_code = unquote_args(code, args)
    r.append([b"qq", unquoted_code])
    r = d2.to(r)
    return r


def parse_mod_sexp(declaration_sexp, namespace, functions, constants, macros, run_program):
    op = declaration_sexp.first().as_atom()
    name = declaration_sexp.rest().first()
    if op == b"include":
        parse_include(name, namespace, functions, constants, macros, run_program)
        return
    name = name.as_atom()
    if name in namespace:
        raise SyntaxError('symbol "%s" redefined' % name.decode())
    namespace.add(name)
    if op == b"defmacro":
        macros.append(declaration_sexp)
    elif op == b"defun":
        functions[name] = declaration_sexp.rest().rest()
    elif op == b"defun-inline":
        macros.append(defun_inline_to_macro(declaration_sexp))
    elif op == b"defconstant":
        constants[name] = declaration_sexp.to(quote(declaration_sexp.rest().rest().first()))
    else:
        raise SyntaxError("expected defun, defun-inline, defmacro, or defconstant")


def compile_mod_stage_1(args, run_program):
    """
    stage 1: collect up names of globals (functions, constants, macros)
    """

    functions = {}
    constants = {}
    macros = []
    main_local_arguments = args.first()

    namespace = set()
    while True:
        args = args.rest()
        if args.rest().nullp():
            break
        parse_mod_sexp(args.first(), namespace, functions, constants, macros, run_program)

    uncompiled_main = args.first()

    functions[MAIN_NAME] = args.to([main_local_arguments, uncompiled_main])

    return functions, constants, macros


def symbol_table_for_tree(tree, root_node):
    if tree.nullp():
        return []

    if not tree.listp():
        return [[tree, root_node.as_path()]]

    left = symbol_table_for_tree(tree.first(), root_node + LEFT)
    right = symbol_table_for_tree(tree.rest(), root_node + RIGHT)

    return left + right


def build_macro_lookup_program(macro_lookup, macros, run_program):
    macro_lookup_program = macro_lookup.to(quote(macro_lookup))
    for macro in macros:
        macro_lookup_program = eval(macro_lookup.to(
            [b"opt", [b"com", quote([CONS_ATOM, macro, macro_lookup_program]), macro_lookup_program]]),
            TOP.as_path())
        macro_lookup_program = optimize_sexp(macro_lookup_program, run_program)
    return macro_lookup_program


def compile_functions(functions, macro_lookup_program, constants_symbol_table, args_root_node):
    compiled_functions = {}
    for name, lambda_expression in functions.items():
        local_symbol_table = symbol_table_for_tree(lambda_expression.first(), args_root_node)
        all_symbols = local_symbol_table + constants_symbol_table
        compiled_functions[name] = lambda_expression.to(
            [b"opt", [b"com",
                      quote(lambda_expression.rest().first()),
                      macro_lookup_program,
                      quote(all_symbols)]])
    return compiled_functions


def compile_mod(args, macro_lookup, symbol_table, run_program):
    """
    Deal with the "mod" keyword.
    """
    (functions, constants, macros) = compile_mod_stage_1(args, run_program)

    # move macros into the macro lookup

    macro_lookup_program = build_macro_lookup_program(macro_lookup, macros, run_program)

    # get a list of all symbols that are possibly used

    all_constants_names = build_used_constants_names(functions, constants, macros)
    has_constants_tree = len(all_constants_names) > 0

    # build defuns table, with function names as keys

    constants_tree = args.to(build_tree(all_constants_names))

    constants_root_node = LEFT
    if has_constants_tree:
        args_root_node = RIGHT
    else:
        args_root_node = TOP

    constants_symbol_table = symbol_table_for_tree(constants_tree, constants_root_node)

    compiled_functions = compile_functions(
        functions, macro_lookup_program, constants_symbol_table, args_root_node)

    main_path_src = binutils.disassemble(compiled_functions[MAIN_NAME])

    if has_constants_tree:
        all_constants_lookup = {k: v for k, v in compiled_functions.items() if k in all_constants_names}
        all_constants_lookup.update(constants)

        all_constants_list = [all_constants_lookup[_] for _ in all_constants_names]
        all_constants_tree_program = args.to(build_tree_program(all_constants_list))

        all_constants_tree_src = binutils.disassemble(all_constants_tree_program)
        arg_tree_src = "(c %s 1)" % all_constants_tree_src
    else:
        arg_tree_src = "1"

    main_code = "(opt (q . (a %s %s)))" % (main_path_src, arg_tree_src)

    if has_constants_tree:
        build_symbol_dump(all_constants_lookup, run_program, "main.sym")

    return binutils.assemble(main_code)



import typing

from klvm import casts, SExp

from .Type import Type, CONS_TYPES

CastableType = typing.Any


def ir_new(type: int, val: CastableType, offset: typing.Optional[int] = None) -> SExp:
    if offset is not None:
        type = SExp.to((type, offset))
    return SExp.to((type, val))


def ir_cons(first: SExp, rest: SExp, offset: typing.Optional[int] = None) -> SExp:
    return ir_new(Type.CONS, ir_new(first, rest), offset)


def ir_list(*items) -> SExp:
    if items:
        return ir_cons(items[0], ir_list(*items[1:]))
    return ir_null()


def ir_null() -> SExp:
    return ir_new(Type.NULL, 0)


def ir_type(ir_sexp: SExp) -> int:
    the_type = ir_sexp.first()
    if the_type.listp():
        the_type = the_type.first()

    return casts.int_from_bytes(the_type.as_atom())


def ir_as_int(ir_sexp: SExp) -> int:
    return casts.int_from_bytes(ir_as_atom(ir_sexp))


def ir_offset(ir_sexp: SExp) -> int:
    the_offset = ir_sexp.first()
    if the_offset.listp():
        the_offset = the_offset.rest().as_atom()
    else:
        the_offset = b"\xff"
    return casts.int_from_bytes(the_offset)


def ir_val(ir_sexp: SExp) -> SExp:
    return ir_sexp.rest()


def ir_nullp(ir_sexp: SExp) -> bool:
    return ir_type(ir_sexp) == Type.NULL


def ir_listp(ir_sexp: SExp) -> bool:
    return ir_type(ir_sexp) in CONS_TYPES


def ir_as_sexp(ir_sexp: SExp) -> SExp:
    if ir_nullp(ir_sexp):
        return []
    if ir_type(ir_sexp) == Type.CONS:
        return ir_as_sexp(ir_first(ir_sexp)).cons(ir_as_sexp(ir_rest(ir_sexp)))
    return ir_sexp.rest()


def ir_is_atom(ir_sexp: SExp) -> bool:
    return not ir_listp(ir_sexp)


def ir_as_atom(ir_sexp: SExp) -> bytes:
    return bytes(ir_sexp.rest().as_atom())


def ir_first(ir_sexp: SExp) -> SExp:
    return ir_sexp.rest().first()


def ir_rest(ir_sexp: SExp) -> SExp:
    return ir_sexp.rest().rest()


def ir_symbol(symbol) -> typing.Tuple[Type, str]:
    return (Type.SYMBOL, symbol.encode("utf8"))


def ir_as_symbol(ir_sexp: SExp) -> typing.Optional[str]:
    if ir_sexp.listp() and ir_type(ir_sexp) == Type.SYMBOL:
        return bytes(ir_as_sexp(ir_sexp).as_atom()).decode("utf8")
    return None


def ir_iter(ir_sexp: SExp) -> typing.Iterable[SExp]:
    while ir_listp(ir_sexp):
        yield ir_first(ir_sexp)
        ir_sexp = ir_rest(ir_sexp)


def is_ir(sexp) -> bool:
    if sexp.atom is not None:
        return False

    type_sexp, val_sexp = sexp.pair
    f = type_sexp.atom
    if f is None or len(f) > 1:
        return False

    the_type = casts.int_from_bytes(f)
    try:
        t = Type(the_type)
    except ValueError:
        return False

    if t == Type.CONS:
        if val_sexp.atom == b"":
            return True
        if val_sexp.pair:
            return all(is_ir(_) for _ in val_sexp.pair)
        return False

    return val_sexp.atom is not None



import enum


def i(b):
    return int.from_bytes(b, "big")


class Type(enum.IntEnum):
    CONS = i(b"CONS")

    NULL = i(b"NULL")
    INT = i(b"INT")
    HEX = i(b"HEX")
    QUOTES = i(b"QT")
    DOUBLE_QUOTE = i(b"DQT")
    SINGLE_QUOTE = i(b"SQT")
    SYMBOL = i(b"SYM")
    OPERATOR = i(b"OP")
    CODE = i(b"CODE")
    NODE = i(b"NODE")

    def listp(self):
        return False

    def as_atom(self):
        return self.to_bytes(len(self), "big", signed=False)

    def __len__(self):
        return (self.bit_length() + 7) >> 3


CONS_TYPES = [Type.CONS]



from klvm import to_sexp_f


from klvm_tools import binutils


"""
"function" is used in front of a constant uncompiled
program to indicate we want this program literal to be
compiled and quoted, so it can be passed as an argument
to a compiled klvm program.

EG: (function (+ 20 @)) should return (+ (q . 20) 1) when run.
Thus (opt (com (q . (function (+ 20 @)))))
should return (q . (+ (q . 20) 1))

(function PROG) => (opt (com (q . PROG) (q . MACROS)))

We have to use "opt" as (com PROG) might leave
some partial "com" operators in there and our
goals is to compile PROG as much as possible.
"""


DEFAULT_MACROS_SRC = [
    """
    ; we have to compile this externally, since it uses itself
    ;(defmacro defmacro (name params body)
    ;    (qq (list (unquote name) (mod (unquote params) (unquote body))))
    ;)
    (q . ("defmacro"
       (c (q . "list")
          (c (f 1)
             (c (c (q . "mod")
                   (c (f (r 1))
                      (c (f (r (r 1)))
                         (q . ()))))
                (q . ()))))))
    """,
    """
    ;(defmacro list ARGS
    ;    ((c (mod args
    ;        (defun compile-list
    ;               (args)
    ;               (if args
    ;                   (qq (c (unquote (f args))
    ;                         (unquote (compile-list (r args)))))
    ;                   ()))
    ;            (compile-list args)
    ;        )
    ;        ARGS
    ;    ))
    ;)
    (q "list"
        (a (q #a (q #a 2 (c 2 (c 3 (q))))
                 (c (q #a (i 5
                             (q #c (q . 4)
                                   (c 9 (c (a 2 (c 2 (c 13 (q))))
                                           (q)))
                             )
                             (q 1))
                           1)
                    1))
            1))
    """,
    """
    (defmacro function (BODY)
        (qq (opt (com (q . (unquote BODY))
                 (qq (unquote (macros)))
                 (qq (unquote (symbols)))))))""",
    """
    (defmacro if (A B C)
        (qq (a
            (i (unquote A)
               (function (unquote B))
               (function (unquote C)))
            @)))""",
    # / operator at the klvm layer is becoming deprecated and
    # will be implemented using divmod.
    """(defmacro / (A B) (qq (f (divmod (unquote A) (unquote B)))))""",
]


DEFAULT_MACRO_LOOKUP = None

def build_default_macro_lookup(eval):
    run = binutils.assemble("(a (com 2 3) 1)")
    global DEFAULT_MACRO_LOOKUP
    for macro_src in DEFAULT_MACROS_SRC:
        macro_sexp = binutils.assemble(macro_src)
        env = macro_sexp.to((macro_sexp, DEFAULT_MACRO_LOOKUP))
        cost, new_macro = eval(run, env)
        DEFAULT_MACRO_LOOKUP = new_macro.cons(
            DEFAULT_MACRO_LOOKUP)
    return DEFAULT_MACRO_LOOKUP


def default_macro_lookup(eval):
    global DEFAULT_MACRO_LOOKUP
    if DEFAULT_MACRO_LOOKUP is None:
        DEFAULT_MACRO_LOOKUP = to_sexp_f([])
        DEFAULT_MACRO_LOOKUP = build_default_macro_lookup(eval)
    return DEFAULT_MACRO_LOOKUP



# read strings into Token

from typing import Iterator, Optional, Tuple

from klvm import to_sexp_f
from klvm.KLVMObject import KLVMObject

from .Type import Type
from .utils import ir_new, ir_cons

Token = Tuple[str, int]
Stream = Iterator[Token]


def consume_whitespace(s: str, offset: int) -> int:
    """
    This also deals with comments.
    """
    while True:
        while offset < len(s) and s[offset].isspace():
            offset += 1
        if offset >= len(s) or s[offset] != ";":
            break
        while offset < len(s) and s[offset] not in "\n\r":
            offset += 1
    return offset


def consume_until_whitespace(s: str, offset: int) -> Token:
    start = offset
    while offset < len(s) and not s[offset].isspace() and s[offset] != ")":
        offset += 1
    return s[start:offset], offset


def next_cons_token(stream: Stream) -> Token:
    for token, offset in stream:
        break
    else:
        raise SyntaxError("missing )")
    return token, offset


def tokenize_cons(token: str, offset: int, stream: Stream) -> KLVMObject:
    if token == ")":
        return ir_new(Type.NULL, 0, offset)

    initial_offset = offset

    first_sexp = tokenize_sexp(token, offset, stream)

    token, offset = next_cons_token(stream)
    if token == ".":
        dot_offset = offset
        # grab the last item
        token, offset = next_cons_token(stream)
        rest_sexp = tokenize_sexp(token, offset, stream)
        token, offset = next_cons_token(stream)
        if token != ")":
            raise SyntaxError("illegal dot expression at %s" % dot_offset)
    else:
        rest_sexp = tokenize_cons(token, offset, stream)
    return ir_cons(first_sexp, rest_sexp, initial_offset)


def tokenize_int(token: str, offset: int) -> Optional[KLVMObject]:
    try:
        return ir_new(Type.INT, int(token), offset)
    except (ValueError, TypeError):
        pass
    return None


def tokenize_hex(token: str, offset: int) -> Optional[KLVMObject]:
    if token[:2].upper() == "0X":
        try:
            token = token[2:]
            if len(token) % 2 == 1:
                token = "0%s" % token
            return ir_new(Type.HEX, bytes.fromhex(token), offset)
        except Exception:
            raise SyntaxError("invalid hex at %s: 0x%s" % (offset, token))
    return None


def tokenize_quotes(token: str, offset: int):
    if len(token) < 2:
        return None
    c = token[:1]
    if c not in "'\"":
        return None

    if token[-1] != c:
        raise SyntaxError("unterminated string starting at %s: %s" % (offset, token))

    q_type = Type.SINGLE_QUOTE if c == "'" else Type.DOUBLE_QUOTE

    return ((q_type, offset), token[1:-1].encode("utf8"))


def tokenize_symbol(token: str, offset: int):
    return ((Type.SYMBOL, offset), token.encode("utf8"))


def tokenize_sexp(token: str, offset: int, stream: Stream):

    if token == "(":
        token, offset = next_cons_token(stream)
        return tokenize_cons(token, offset, stream)

    for f in [
        tokenize_int,
        tokenize_hex,
        tokenize_quotes,
        tokenize_symbol,
    ]:
        r = f(token, offset)
        if r is not None:
            return r


def token_stream(s: str) -> Stream:
    offset = 0
    while offset < len(s):
        offset = consume_whitespace(s, offset)
        if offset >= len(s):
            break
        c = s[offset]
        if c in "(.)":
            yield c, offset
            offset += 1
            continue
        if c in "\"'":
            start = offset
            initial_c = s[start]
            offset += 1
            while offset < len(s) and s[offset] != initial_c:
                offset += 1
            if offset < len(s):
                yield s[start:offset + 1], start
                offset += 1
                continue
            else:
                raise SyntaxError(
                    "unterminated string starting at %s: %s" % (start, s[start:])
                )
        token, end_offset = consume_until_whitespace(s, offset)
        yield token, offset
        offset = end_offset


def read_ir(s: str, to_sexp=to_sexp_f):
    stream = token_stream(s)

    for token, offset in stream:
        return to_sexp(tokenize_sexp(token, offset, stream))
    else:
        raise SyntaxError("unexpected end of stream")



import string
from typing import Dict

from klvm import KEYWORD_FROM_ATOM, KEYWORD_TO_ATOM, SExp
from klvm.casts import int_from_bytes, int_to_bytes
from klvm.SExp import SExp

from ir.reader import read_ir
from ir.writer import write_ir
from ir.utils import (
    ir_as_symbol, ir_cons, ir_first, ir_listp, ir_null,
    ir_nullp, ir_rest, ir_symbol, ir_val, is_ir
)
from ir.Type import Type


def assemble_from_ir(ir_sexp):
    keyword = ir_as_symbol(ir_sexp)
    if keyword:
        if keyword[:1] == "#":
            keyword = keyword[1:]
        atom = KEYWORD_TO_ATOM.get(keyword)
        if atom is not None:
            return ir_sexp.to(atom)
        if True:
            return ir_val(ir_sexp)
        raise SyntaxError(
            "can't parse %s at %s" % (keyword, ir_sexp._offset))

    if not ir_listp(ir_sexp):
        return ir_val(ir_sexp)

    if ir_nullp(ir_sexp):
        return ir_sexp.to([])

    # handle "q"
    first = ir_first(ir_sexp)
    keyword = ir_as_symbol(first)
    if keyword == "q":
        pass
        # TODO: note that any symbol is legal after this point

    sexp_1 = assemble_from_ir(first)
    sexp_2 = assemble_from_ir(ir_rest(ir_sexp))
    return sexp_1.cons(sexp_2)


printable_chars = (
    "0123456789"
    + "abcdefghijklmnopqrstuvwxyz"
    + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    + "!#$%&'()*+,-./:;<=>?@[]^_`{|}~ "
    + "\\"
)


def type_for_atom(atom) -> Type:
    if len(atom) > 2:
        try:
            v = bytes(atom).decode("utf8")
            if all(c in printable_chars for c in v):
                return Type.QUOTES
        except UnicodeDecodeError:
            pass
        return Type.HEX
    if int_to_bytes(int_from_bytes(atom)) == atom:
        return Type.INT
    return Type.HEX


def disassemble_to_ir(sexp, keyword_from_atom, allow_keyword=None):
    if is_ir(sexp) and allow_keyword is not False:
        return ir_cons(ir_symbol("ir"), sexp)

    if sexp.listp():
        if sexp.first().listp() or allow_keyword is None:
            allow_keyword = True
        v0 = disassemble_to_ir(sexp.first(), keyword_from_atom, allow_keyword=allow_keyword)
        v1 = disassemble_to_ir(sexp.rest(), keyword_from_atom, allow_keyword=False)
        return ir_cons(v0, v1)

    as_atom = sexp.as_atom()
    if allow_keyword:
        v = keyword_from_atom.get(as_atom)
        if v is not None and v != '.':
            return ir_symbol(v)

    if sexp.nullp():
        return ir_null()

    return sexp.to((type_for_atom(as_atom), as_atom))


def disassemble(sexp: SExp, keyword_from_atom: Dict[bytes, str] = KEYWORD_FROM_ATOM) -> str:
    symbols = disassemble_to_ir(sexp, keyword_from_atom=keyword_from_atom)
    return write_ir(symbols)


def assemble(s: str) -> SExp:
    symbols = read_ir(s)
    return assemble_from_ir(symbols)



import random
import os
import sys
from enum import Enum

def make_small_value():
    # we use small values as divisors, so it can't be 0
    return '0x%02x' % random.randint(1, 255)

def make_string():
    length = random.randint(5, 10)
    ret = '"'
    for i in range(length):
        ret += chr(random.randint(ord('A'), ord('Z')))
    ret += '"'
    return ret

def make_value():
    r = random.randint(0, 5)
    if r == 0: return make_string()
    length = random.randint(1, 10)
    ret = '0x'
    for i in range(length):
        ret += '%02x' % random.getrandbits(8)
    return ret

def make_point():
        return random.choice([
'0xb3b8ac537f4fd6bde9b26221d49b54b17a506be147347dae5d081c0a6572b611d8484e338f3432971a9823976c6a232b',
'0x8a9970d47a86db8a2261954a6876081722e347912875a96220db69c23be8b48ac2563499f2a4dc9c50fe169615a77d80',
'0xa7e04cdd823ee4d767e99d72f9f73f5795faaf8161ddef3c3433c11793a27526a1909f8618edea2f01899536dca43d90',
'0x92f552fab78a3c4a0f13297b96c56a7ae83dace9569ec65b94c7dc82490091803f01df688629e998894d8a195cba5c23'])

def make_tree(depth):
    if depth == 1:
        return make_value()
    else:
        return '(' + make_tree(depth - 1) + ' . ' + make_tree(depth - 1) + ')'

def make_path(depth):
    ret = 1
    while depth > 1:
        ret <<= 1
        ret |= random.getrandbits(1)
        depth -= 1
    return '%d' % ret

class ExpType(Enum):
    VALUE = 0
    SMALL_VALUE = 1
    LIST = 2
    VALUE_LIST = 3
    PROGRAM = 4
    POINT = 5
    STRING = 6
    ANYTHING = 7

def make_condition(depth, env_depth):
    r = random.randint(0, 5)

    if r == 3:
        return make_expression(depth - 1, env_depth, ExpType.VALUE)
    elif r == 4:
        return '(l ' + make_expression(depth -1, env_depth) + ')'
    elif r == 5:
        return '(q . ' + make_small_value() + ')'
    elif r == 0:
        ret = '(= '
    elif r == 1:
        ret = '(> '
    elif r == 2:
        ret = '(>s '

    ret += make_expression(depth - 1, env_depth, ExpType.VALUE) + \
        ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ')'
    return ret

def make_expression(depth, env_depth, kind = ExpType.ANYTHING):
    if depth <= 0:
        if kind == ExpType.SMALL_VALUE:
            return '(q . ' + make_small_value() + ')'
        elif kind == ExpType.VALUE:
            return '(q . ' + make_value() + ')'
        elif kind == ExpType.LIST or kind == ExpType.VALUE_LIST or kind == ExpType.ANYTHING:
            return '(q . (' + make_value() + ' ' + make_value() + '))'
        elif kind == ExpType.POINT:
            return '(q . ' + make_point() + ')'
        elif kind == ExpType.STRING:
            return '(q . ' + make_string() + ')'
        else:
            return '(q . (q . ' + make_value() + '))'

    value_kind = ExpType.VALUE

    if kind == ExpType.VALUE:
        r = random.choice([0, 1, 3, 5, 25, 26, 11, 12, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38])
    elif kind == ExpType.SMALL_VALUE:
        value_kind = ExpType.SMALL_VALUE
        r = random.choice([5, 25, 26, 11, 12, 17, 18, 19, 20, 23, 24, 28, 35, 36, 37, 28])
    elif kind == ExpType.POINT:
        r = random.choice([5, 25, 26, 29, 31, 31])
    elif kind == ExpType.STRING:
        r = random.choice([5, 16, 25, 26, 33])
    elif kind == ExpType.LIST:
        r = random.choice([2, 5, 25, 26, 7, 8, 13])
    elif kind == ExpType.VALUE_LIST:
        r = random.choice([5, 25, 26, 7, 8, 13])
    elif kind == ExpType.PROGRAM:
        r = random.choice([5, 25, 26, 9])
    else:
        r = random.randint(0, 38)

    if r == 0:
        # value
        return '(q . ' + make_value() + ')'
    elif r == 1:
        # path
        return make_path(env_depth)
    elif r == 2:
        # cons
        return '(c ' + make_expression(depth - 1, env_depth) + ' ' + make_expression(depth - 1, env_depth, ExpType.LIST) + ')'
    elif r == 3:
        # first (resulting in a value)
        return '(f ' + make_expression(depth - 1, env_depth, ExpType.VALUE_LIST) + ')'
    elif r == 4:
        # rest
        return '(r ' + make_expression(depth - 1, env_depth, ExpType.LIST) + ')'
    elif r == 5 or r == 25 or r == 26:
        # if-condition
        return '(i ' + make_condition(depth - 1, env_depth) + \
            ' ' + make_expression(depth - 1, env_depth, kind) + \
            ' ' + make_expression(depth - 1, env_depth, kind) + ')'
    elif r == 6:
        # first (resulting in anything)
        return '(f ' + make_expression(depth - 1, env_depth, ExpType.LIST) + ')'
    elif r == 7:
        # cons (resulting in a value list)
        return '(c ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + \
            ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE_LIST) + ')'
    elif r == 8:
        # value list
        return '(q . (' + make_value() + ' ' + make_value() + '))'
    elif r == 9 or r == 10:
        # eval
        new_env_depth = random.randint(1, 5)
        return '(a (q . ' + make_expression(depth - 1, new_env_depth) + ') (q . ' + make_tree(new_env_depth) + '))'
    elif r == 11:
        # add expression
        ret = '(+'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, value_kind)
        return ret + ')'
    elif r == 12:
        # subtract expression
        ret = '(-'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, value_kind)
        return ret + ')'
    elif r == 13:
        # divmod expression
        return '(divmod ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ' (q . ' + \
            make_small_value() + '))'
    elif r == 14:
        # divmod expression, integer division
        return '(f (divmod ' + make_expression(depth - 1, env_depth, value_kind) + ' (q . ' + \
            make_small_value() + ')))'
    elif r == 15:
        # divmod expression, remainder
        return '(r (divmod ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ' (q . ' + \
            make_small_value() + ')))'
    elif r == 16:
        # concat expression
        ret = '(concat'
        n = random.randint(1, 4)
        for i in range(n):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        ret += ' (q . ' + make_string() + ')'
        return ret + ')'
    elif r == 17:
        # logical or
        ret = '(logior'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, value_kind)
        return ret + ')'
    elif r == 18:
        # logical xor
        ret = '(logxor'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, value_kind)
        return ret + ')'
    elif r == 19:
        # logical and
        ret = '(logand'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, value_kind)
        return ret + ')'
    elif r == 20:
        # logical not
        return '(lognot ' + make_expression(depth - 1, env_depth, value_kind) + ')'
    elif r == 21:
        # arithmetic shift
        return '(ash ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ' ' + \
            make_expression(depth - 1, env_depth, ExpType.SMALL_VALUE) + ')'
    elif r == 22:
        # left shift
        return '(lsh ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ' ' + \
            make_expression(depth - 1, env_depth, ExpType.SMALL_VALUE) + ')'
    elif r == 23:
        # strlen
        return '(strlen ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ')'
    elif r == 24:
        return '(q . ' + make_small_value() + ')'
    elif r == 27:
        # multiply expression
        ret = '(*'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        return ret + ')'
    elif r == 28:
        # listp
        return '(l ' + make_expression(depth - 1, env_depth) + ')'
    elif r == 29:
        # pubkey_for_exp
        return '(pubkey_for_exp ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ')'
    elif r == 30:
        # a valid point
       return '(q . ' + make_point() + ')'
    elif r == 31:
        # point_add
        ret = '(point_add'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.POINT)
        return ret + ')'
    elif r == 32:
        # substring
        return '(substr ' + make_expression(depth - 1, env_depth, ExpType.STRING) + ' (q . 0) (q . 1))';
    elif r == 33:
        return '(q . ' + make_string() + ')'
    elif r == 34:
        ret = '(sha256'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        return ret + ')'
    elif r == 35:
        ret = '(any'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        return ret + ')'
    elif r == 36:
        ret = '(all'
        for i in range(random.randint(0, 4)):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        return ret + ')'
    elif r == 37:
        return '(not ' + make_expression(depth - 1, env_depth, ExpType.VALUE) + ')'
    elif r == 38:
        # concat expression (may return an empty string)
        ret = '(concat'
        n = random.randint(0, 4)
        for i in range(n):
            ret += ' ' + make_expression(depth - 1, env_depth, ExpType.VALUE)
        return ret + ')'

    print('ERROR: r=%d' % r)

left_to_generate = 1000

try: os.mkdir('mixed-programs')
except: pass

if len(sys.argv) > 1:
    seed = int(sys.argv[1])
    left_to_generate = 1
else:
    seed = random.getrandbits(32)

while left_to_generate > 0:
    left_to_generate -= 1
    random.seed(seed)

    env_depth = random.randint(1, 10)
    depth = random.randint(16, 20)

    print('seed:', seed)

    program = make_expression(depth, env_depth)
    arguments = make_tree(env_depth)
    filename = 'mixed-programs/mixed-%08x.klvm' % seed
    env_filename = 'mixed-programs/mixed-%08x.env' % seed
    prg = open(filename, 'w+').write(program)
    env_file = open(env_filename, 'w+').write(arguments)

    seed = random.getrandbits(32)



import io
import unittest

from klvm import to_sexp_f
from klvm.serialize import sexp_from_stream


TEXT = b"the quick brown fox jumps over the lazy dogs"


class SerializeTest(unittest.TestCase):
    def check_serde(self, s):
        v = to_sexp_f(s)
        b = v.as_bin()
        v1 = sexp_from_stream(io.BytesIO(b), to_sexp_f)
        if v != v1:
            print("%s: %d %s %s" % (v, len(b), b, v1))
            breakpoint()
            b = v.as_bin()
            v1 = sexp_from_stream(io.BytesIO(b), to_sexp_f)
        self.assertEqual(v, v1)

    def test_empty_string(self):
        self.check_serde(b"")

    def test_single_bytes(self):
        for _ in range(256):
            self.check_serde(bytes([_]))

    def test_short_lists(self):
        self.check_serde([])
        for _ in range(0, 2048, 8):
            for size in range(1, 5):
                self.check_serde([_] * size)

    def test_cons_box(self):
        self.check_serde((None, None))
        self.check_serde((None, [1, 2, 30, 40, 600, (None, 18)]))
        self.check_serde((100, (TEXT, (30, (50, (90, (TEXT, TEXT + TEXT)))))))

    def test_long_blobs(self):
        text = TEXT * 300
        for _, t in enumerate(text):
            t1 = text[:_]
            self.check_serde(t1)

    def test_very_long_blobs(self):
        for size in [0x40, 0x2000, 0x100000, 0x8000000]:
            count = size // len(TEXT)
            text = TEXT * count
            assert len(text) < size
            self.check_serde(text)
            text = TEXT * (count + 1)
            assert len(text) > size
            self.check_serde(text)



r"""
We treat an s-expression as a binary tree, where leaf nodes are atoms and pairs
are nodes with two children. We then number the paths as follows:

              1
             / \
            /   \
           /     \
          /       \
         /         \
        /           \
       2             3
      / \           / \
     /   \         /   \
    4      6      5     7
   / \    / \    / \   / \
  8   12 10  14 9  13 11  15

etc.

You're probably thinking "the first two rows make sense, but why do the numbers
do that weird thing after?" The reason has to do with making the implementation simple.
We want a simple loop which starts with the root node, then processes bits starting with
the least significant, moving either left or right (first or rest). So the LEAST significant
bit controls the first branch, then the next-least the second, and so on. That leads to this
ugly-numbered tree.
"""


def compose_paths(path_0, path_1):
    """
    The binary representation of a path is a 1 (which means "stop"), followed by the
    path as binary digits, where 0 is "left" and 1 is "right".

    Look at the diagram at the top for these examples.

    Example: 9 = 0b1001, so right, left, left
    Example: 10 = 0b1010, so left, right, left

    How it works: we write both numbers as binary. We ignore the terminal in path_0, since it's
    not the terminating condition anymore. We shift path_1 enough places to OR in the rest of path_0.

    Example: path_0 = 9 = 0b1001, path_1 = 10 = 0b1010.
    Shift path_1 three places (so there is room for 0b001) to 0b1010000.
    Then OR in 0b001 to yield 0b1010001 = 81, which is right, left, left, left, right, left.
    """
    mask = 1
    temp_path = path_0
    while temp_path > 1:
        path_1 <<= 1
        mask <<= 1
        temp_path >>= 1

    mask -= 1
    path = path_1 | (path_0 & mask)
    return path


class NodePath:
    """
    Use 1-based paths
    """

    def __init__(self, index=1):
        if index < 0:
            byte_count = (index.bit_length() + 7) >> 3
            blob = index.to_bytes(byte_count, byteorder="big", signed=True)
            index = int.from_bytes((b"\0" + blob), byteorder="big", signed=False)
        self._index = index

    def as_short_path(self):
        index = self._index
        byte_count = (index.bit_length() + 7) >> 3
        return index.to_bytes(byte_count, byteorder="big")

    as_path = as_short_path

    def __add__(self, other_node):
        return self.__class__(compose_paths(self._index, other_node._index))

    def first(self):
        return self.__class__(self._index * 2)

    def rest(self):
        return self.__class__(self._index * 2 + 1)

    def __str__(self):
        return "NodePath: %d" % self._index

    def __repr__(self):
        return "NodePath: %d" % self._index


TOP = NodePath()
LEFT = TOP.first()
RIGHT = TOP.rest()



from klvm_tools.binutils import assemble
from klvm_tools.pattern_match import match


def test_pattern_match():

    r = match(assemble("($ . $)"), assemble("$"))
    assert r == {}

    r = match(assemble("($ . $)"), assemble("x"))
    assert r is None

    r = match(assemble("(: . :)"), assemble(":"))
    assert r == {}

    r = match(assemble("(: . :)"), assemble("x"))
    assert r is None

    r = match(assemble("$"), assemble("$"))
    assert r == {}

    # () is an atom
    r = match(assemble("($ . n)"), assemble("()"))
    assert r == {"n": assemble("()")}

    r = match(assemble("($ . size)"), assemble("200"))
    assert r == {"size": assemble("200")}

    r = match(assemble("(: . size)"), assemble("200"))
    assert r == {"size": assemble("200")}

    r = match(assemble("($ . size)"), assemble("(I like cheese)"))
    assert r is None

    r = match(assemble("(: . size)"), assemble("(I like cheese)"))
    assert r == {"size": assemble("(I like cheese)")}

    r = match(
        assemble("(= (f (r (a))) ($ . pubkey))"), assemble("(= (f (r (a))) 50000)")
    )
    assert r == {"pubkey": assemble("50000")}

    r = match(
        assemble("(= (f (r (a))) ($ . pubkey1) ($ . pubkey2))"),
        assemble("(= (f (r (a))) 50000 60000)"),
    )
    assert r == {"pubkey1": assemble("50000"), "pubkey2": assemble("60000")}

    r = match(
        assemble("(= (f (r (a))) ($ . pubkey1) ($ . pubkey1))"),
        assemble("(= (f (r (a))) 50000 60000)"),
    )
    assert r is None

    r = match(
        assemble("(= (f (r (a))) ($ . pubkey1) ($ . pubkey1))"),
        assemble("(= (f (r (a))) 50000 50000)"),
    )
    assert r == {"pubkey1": assemble("50000")}



import hashlib


def sha256tree(v):
    pair = v.pair
    if pair:
        left = sha256tree(pair[0])
        right = sha256tree(pair[1])
        s = b"\2" + left + right
    else:
        s = b"\1" + v.atom
    return hashlib.sha256(s).digest()



import glob
import subprocess
import sys
import os

results = open('mixed-programs/results.csv', 'w+')

gnuplot_filename = 'mixed-programs/render.gnuplot'
gnuplot = open(gnuplot_filename, 'w+')

gnuplot.write('''set output "mixed-programs/costs.png"
set datafile separator ","
set term png size 1400,900 small
set termoption enhanced
set ylabel "run-time (s)"
set xlabel "cost"
set xrange [0:*]
set yrange [0:*]
plot "mixed-programs/results.csv" using 1:2 with points
''')
gnuplot.close()

counter = 0
for fn in glob.glob('mixed-programs/*.klvm'):

    print('%04d: %s' % (counter, fn))
    counter += 1
    # the filename is expected to be in the form:
    # name "-" value_size "-" num_calls
    env = open(fn[:-4] + 'env').read()
    output = subprocess.check_output(['brun', '--backend=rust', '-c', '--quiet', '--time', fn, env])
    output = output.decode('ascii').split('\n', 5)[:-1]

    counters = {}
    for o in output:
        if ':' in o:
            key, value = o.split(':')
            counters[key.strip()] = value.strip()
        elif '=' in o:
            key, value = o.split('=')
            counters[key.strip()] = value.strip()

    line = counters['cost'] + ',' + counters['run_program'] + '\n'
    results.write(line)
    results.flush()

results.close()
os.system('gnuplot %s' % gnuplot_filename)



import hashlib
import json

from typing import Any, Callable, List

from klvm import SExp

from .sha256tree import sha256tree

OpCallable = Callable[[Any, "ValStackType"], int]

ValStackType = List[SExp]
OpStackType = List[OpCallable]


PRELUDE = '''<html>
<head>
  <link rel="stylesheet"
      href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css"
      integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T"
      crossorigin="anonymous">
  <script
      src="https://code.jquery.com/jquery-3.3.1.slim.min.js"
      integrity="sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo"
      crossorigin="anonymous"></script>
  <script
      src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js"
      integrity="sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1"
      crossorigin="anonymous"></script>
  <script
      src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js"
      integrity="sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM"
      crossorigin="anonymous"></script>
</head>
<body>
<div class="container">
'''


TRAILER = "</div></body></html>"


def dump_sexp(s, disassemble):
    return '<span id="%s">%s</span>' % (id(s), disassemble(s))


def dump_invocation(form, rewrit_form, env, result, disassemble):
    print('<hr><div class="invocation" id="%s">' % id(form))
    print('<span class="form"><a name="id_%s">%s</a></span>' % (
        id(form), dump_sexp(form, disassemble)))
    print('<ul>')
    if form != rewrit_form:
        print(
            '<li>Rewritten as:<span class="form">'
            '<a name="id_%s">%s</a></span></li>' % (
                id(rewrit_form), dump_sexp(rewrit_form, disassemble)))
    for _, e in enumerate(env):
        print('<li>x%d: <a href="#id_%s">%s</a></li>' % (
            _, id(e), dump_sexp(e, disassemble)))
    print('</ul>')
    print('<span class="form">%s</span>' % dump_sexp(result, disassemble))
    if form.listp() and len(form) > 1:
        print('<ul>')
        for _, arg in enumerate(form[1:]):
            print('<li>arg %d: <a href="#id_%s">%s</a></li>' % (
                _, id(arg), dump_sexp(arg, disassemble)))
        print('</ul>')
    print("</div>")


def trace_to_html(invocations, disassemble):
    invocations = reversed(invocations)

    print(PRELUDE)

    id_set = set()
    id_list = []

    for form, rewrit_form, env, rv in invocations:
        dump_invocation(form, rewrit_form, env, rv, disassemble)
        the_id = id(form)
        if the_id not in id_set:
            id_set.add(the_id)
            id_list.append(form)

    print('<hr>')
    for _ in id_list:
        print('<div><a href="#id_%s">%s</a></div>' % (id(_), disassemble(_)))
    print('<hr>')

    print(TRAILER)


def build_symbol_dump(constants_lookup, run_program, path):
    compiled_lookup = {}
    for k, v in constants_lookup.items():
        cost, v1 = run_program(v, v.null())
        compiled_lookup[sha256tree(v1).hex()] = bytes(k).decode()
    output = json.dumps(compiled_lookup)
    with open(path, "w") as f:
        f.write(output)


def text_trace(disassemble, form, symbol, env, result):
    if symbol:
        env = env.rest()
        symbol = disassemble(env.to(symbol.encode()).cons(env))
    else:
        symbol = "%s [%s]" % (disassemble(form), disassemble(env))
    print("%s => %s" % (symbol, result))
    print("")


def table_trace(disassemble, form, symbol, env, result):
    if form.listp():
        sexp = form.first()
        args = form.rest()
    else:
        sexp = form
        args = form.null()
    print("exp:", disassemble(sexp))
    print("arg:", disassemble(args))
    print("env:", disassemble(env))
    print("val:", result)
    print("bexp:", sexp.as_bin())
    print("barg:", args.as_bin())
    print("benv:", env.as_bin())
    print("--")


def display_trace(trace, disassemble, symbol_table, display_fun):
    for item in trace:
        form, env, rv = item
        if rv is None:
            rv = "(didn't finish)"
        else:
            rv = disassemble(rv)
        h = sha256tree(form).hex()
        symbol = symbol_table.get(h) if symbol_table else symbol_table
        display_fun(disassemble, form, symbol, env, rv)


def trace_to_text(trace, disassemble, symbol_table):
    display_trace(trace, disassemble, symbol_table, text_trace)


def trace_to_table(trace, disassemble, symbol_table):
    display_trace(trace, disassemble, symbol_table, table_trace)


def make_trace_pre_eval(log_entries, symbol_table=None):
    def pre_eval_f(sexp, args):
        sexp, args = [SExp.to(_) for _ in [sexp, args]]
        if symbol_table:
            h = sha256tree(sexp).hex()
            if h not in symbol_table:
                return None
        log_entry = [sexp, args, None]
        log_entries.append(log_entry)

        def callback_f(r):
            log_entry[-1] = SExp.to(r)

        return callback_f

    return pre_eval_f



# read strings into Token

import io
from typing import Iterator

from klvm import casts, SExp
from klvm.serialize import sexp_to_stream

from .Type import Type
from .utils import ir_nullp, ir_type, ir_listp, ir_first, ir_rest, ir_as_atom, ir_val


def iter_sexp_format(ir_sexp: SExp) -> Iterator[str]:
    yield "("
    is_first = True
    while not ir_nullp(ir_sexp):
        if not ir_listp(ir_sexp):
            yield " . "
            yield from iter_ir_format(ir_sexp)
            break
        if not is_first:
            yield " "
        for _ in iter_ir_format(ir_first(ir_sexp)):
            yield _
        ir_sexp = ir_rest(ir_sexp)
        is_first = False
    yield ")"


def iter_ir_format(ir_sexp: SExp) -> Iterator[str]:
    if ir_listp(ir_sexp):
        yield from iter_sexp_format(ir_sexp)
        return

    type = ir_type(ir_sexp)

    if type == Type.CODE:
        bio = io.BytesIO()
        sexp_to_stream(ir_val(ir_sexp), bio)
        code = bio.getvalue().hex()
        yield "CODE[%s]" % code
        return

    if type == Type.NULL:
        yield "()"
        return

    atom = ir_as_atom(ir_sexp)

    if type == Type.INT:
        yield "%d" % casts.int_from_bytes(atom)
    elif type == Type.NODE:
        yield "NODE[%d]" % casts.int_from_bytes(atom)
    elif type == Type.HEX:
        yield "0x%s" % atom.hex()
    elif type == Type.QUOTES:
        yield '"%s"' % atom.decode("utf8")
    elif type == Type.DOUBLE_QUOTE:
        yield '"%s"' % atom.decode("utf8")
    elif type == Type.SINGLE_QUOTE:
        yield "'%s'" % atom.decode("utf8")
    elif type in (Type.SYMBOL, Type.OPERATOR):
        try:
            yield atom.decode("utf8")
        except UnicodeDecodeError:
            yield "(indecipherable symbol: %s)" % atom.hex()
    else:
        raise SyntaxError("bad ir format: %s" % ir_sexp)


def write_ir_to_stream(ir_sexp: SExp, f):
    for _ in iter_ir_format(ir_sexp):
        f.write(_)


def write_ir(ir_sexp: SExp) -> str:
    s = io.StringIO()
    write_ir_to_stream(ir_sexp, s)
    return s.getvalue()



import pathlib

from ir.reader import read_ir
from ir.writer import write_ir_to_stream


from klvm.EvalError import EvalError
from klvm_tools.operator_dict import OperatorDict

from klvm_tools.binutils import assemble_from_ir, disassemble_to_ir

from stages.stage_0 import (
    run_program as run_program_0,
    OPERATOR_LOOKUP as ORIGINAL_OPERATOR_LOOKUP,
)

from .compile import make_do_com
from .optimize import make_do_opt


def do_read(args):
    filename = args.first().as_atom()
    s = open(filename).read()
    ir_sexp = args.to(read_ir(s))
    sexp = assemble_from_ir(ir_sexp)
    return 1, sexp


def do_write(args):
    filename = args.first().as_atom()
    data = args.rest().first()
    with open(filename, "w") as f:
        write_ir_to_stream(disassemble_to_ir(data), f)
    return 1, args.to(0)


def run_program_for_search_paths(search_paths):

    def do_full_path_for_name(args):
        filename = args.first().as_atom()
        for path in search_paths:
            f_path = pathlib.Path(path) / bytes(filename).decode()
            if f_path.is_file():
                return 1, args.to(str(f_path).encode())
        raise EvalError("can't open %s" % filename, args)

    operator_lookup = OperatorDict(ORIGINAL_OPERATOR_LOOKUP)

    def run_program(
        program, args, operator_lookup=operator_lookup, max_cost=None, pre_eval_f=None, strict=False
    ):
        return run_program_0(
            program,
            args,
            operator_lookup=operator_lookup,
            max_cost=max_cost,
            pre_eval_f=pre_eval_f,
            strict=strict
        )

    BINDINGS = {
        b"com": make_do_com(run_program),
        b"opt": make_do_opt(run_program),
        b"_full_path_for_name": do_full_path_for_name,
        b"_read": do_read,
        b"_write": do_write,
    }

    operator_lookup.update(BINDINGS)

    return run_program



import argparse
import importlib
import io
import json
import pathlib
import sys
import time

from klvm import to_sexp_f, KEYWORD_FROM_ATOM, KEYWORD_TO_ATOM, SExp
import klvm_tools_rs
from klvm.EvalError import EvalError
from klvm.serialize import sexp_from_stream, sexp_to_stream
from klvm.operators import OP_REWRITE

from ir import reader

from . import binutils
from .debug import make_trace_pre_eval, trace_to_text, trace_to_table
from .sha256tree import sha256tree

try:
    from klvm_rs import run_serialized_chik_program, MEMPOOL_MODE
except ImportError:
    run_serialized_chik_program = None


def path_or_code(arg):
    try:
        with open(arg) as f:
            return f.read()
    except IOError:
        return arg


def stream_to_bin(write_f):
    b = io.BytesIO()
    write_f(b)
    return b.getvalue()


def call_tool(tool_name, desc, conversion, input_args):
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "-H", "--script-hash", action="store_true", help="Show only sha256 tree hash of program"
    )
    parser.add_argument(
        "path_or_code",
        nargs="*",
        type=path_or_code,
        help="path to klvm script, or literal script",
    )

    sys.setrecursionlimit(20000)
    args = parser.parse_args(args=input_args[1:])

    for program in args.path_or_code:
        if program == "-":
            program = sys.stdin.read()
        sexp, text = conversion(program)
        if args.script_hash:
            print(sha256tree(sexp).hex())
        elif text:
            print(text)


def opc(args=sys.argv):
    def conversion(text):
        try:
            ir_sexp = reader.read_ir(text)
            sexp = binutils.assemble_from_ir(ir_sexp)
        except SyntaxError as ex:
            print("%s" % ex.msg)
            return None, None
        return sexp, sexp.as_bin().hex()

    call_tool("opc", "Compile a klvm script.", conversion, args)


def opd(args=sys.argv):
    def conversion(blob):
        sexp = sexp_from_stream(io.BytesIO(bytes.fromhex(blob)), to_sexp_f)
        return sexp, binutils.disassemble(sexp)
    call_tool("opd", "Disassemble a compiled klvm script from hex.", conversion, args)


def stage_import(stage):
    stage_path = "stages.stage_%s" % stage
    try:
        return importlib.import_module(stage_path)
    except ImportError:
        raise ValueError("bad stage: %s" % stage)


def as_bin(streamer_f):
    f = io.BytesIO()
    streamer_f(f)
    return f.getvalue()


def run(args=sys.argv):
    sys.stdout.write(bytes(klvm_tools_rs.launch_tool("run", args, 2)).decode('utf8'))


def brun(args=sys.argv):
    return launch_tool(args, "brun")


def launch_tool(args, tool_name, default_stage=0):
    sys.setrecursionlimit(20000)
    parser = argparse.ArgumentParser(
        description='Execute a klvm script.'
    )
    parser.add_argument(
        "--strict", action="store_true",
        help="deprecated alias for --mempool")
    parser.add_argument(
        "--mempool", action="store_true",
        help="Unknown opcodes are always fatal errors in mempool mode")
    parser.add_argument(
        "-x", "--hex", action="store_true",
        help="Read program and environment as hexadecimal bytecode")
    parser.add_argument(
        "-s", "--stage", type=stage_import,
        help="stage number to include", default=stage_import(default_stage))
    parser.add_argument(
        "-v", "--verbose", action="store_true",
        help="Display resolve of all reductions, for debugging")
    parser.add_argument(
        "-t", "--table", action="store_true",
        help="Print diagnostic table of reductions, for debugging")
    parser.add_argument(
        "-c", "--cost", action="store_true", help="Show cost")
    parser.add_argument(
        "--time", action="store_true", help="Print execution time")
    parser.add_argument(
        "-m", "--max-cost", type=int, default=11000000000, help="Maximum cost")
    parser.add_argument(
        "-d", "--dump", action="store_true",
        help="dump hex version of final output")
    parser.add_argument(
        "--quiet", action="store_true", help="Suppress printing the program result")
    parser.add_argument(
        "-y", "--symbol-table", type=pathlib.Path,
        help=".SYM file generated by compiler")
    parser.add_argument(
        "-n", "--no-keywords", action="store_true",
        help="Output result as data, not as a program")
    parser.add_argument("--backend", type=str, help="force use of 'rust' or 'python' backend")
    parser.add_argument(
        "-i",
        "--include",
        type=pathlib.Path,
        help="add a search path for included files",
        action="append",
        default=[],
    )
    parser.add_argument(
        "path_or_code", type=path_or_code,
        help="filepath to klvm script, or a literal script")

    parser.add_argument(
        "env", nargs="?", type=path_or_code,
        help="klvm script environment, as klvm src, or hex")

    args = parser.parse_args(args=args[1:])

    keywords = {} if args.no_keywords else KEYWORD_FROM_ATOM

    if hasattr(args.stage, "run_program_for_search_paths"):
        run_program = args.stage.run_program_for_search_paths(args.include)
    else:
        run_program = args.stage.run_program

    program_serialized = None
    arg_serialized = None

    time_start = time.perf_counter()
    if args.hex:
        program_serialized = bytes.fromhex(args.path_or_code)
        if not args.env:
            args.env = "80"
        arg_serialized = bytes.fromhex(args.env)
        time_read_hex = time.perf_counter()

    else:

        src_text = args.path_or_code
        try:
            src_sexp = reader.read_ir(src_text)
        except SyntaxError as ex:
            print("FAIL: %s" % (ex))
            return -1
        program_serialized = binutils.assemble_from_ir(src_sexp).as_bin()
        if not args.env:
            args.env = "()"
        env_ir = reader.read_ir(args.env)
        arg_serialized = binutils.assemble_from_ir(env_ir).as_bin()

        time_assemble = time.perf_counter()

    pre_eval_f = None
    symbol_table = None

    log_entries = []

    if args.symbol_table:
        with open(args.symbol_table) as f:
            symbol_table = json.load(f)
        pre_eval_f = make_trace_pre_eval(log_entries, symbol_table)
    elif args.verbose or args.table:
        pre_eval_f = make_trace_pre_eval(log_entries)

    if hasattr(args.stage, tool_name):
        arg_serialized = b"\xff" + program_serialized + arg_serialized
        program_serialized = getattr(args.stage, tool_name).as_bin()

    cost = 0
    try:
        output = "(didn't finish)"

        if args.backend == "rust":
            use_rust = True
        elif args.backend == "python":
            use_rust = False
        else:
            use_rust = (
                tool_name != "run"
                and not pre_eval_f
                and run_serialized_chik_program
                and args.stage.__name__ == "stages.stage_0"
            )

        max_cost = args.max_cost
        if use_rust:
            time_parse_input = time.perf_counter()

            try:
                cost, result = run_serialized_chik_program(
                    program_serialized,
                    arg_serialized,
                    max_cost,
                    MEMPOOL_MODE if (args.mempool or args.strict) else 0,
                )
            except ValueError as ve:
                err = EvalError(ve.args[0], ve.args[1])
                raise err
            time_done = time.perf_counter()
            result = SExp.to(result)
        else:
            program = sexp_from_stream(io.BytesIO(program_serialized), to_sexp_f)
            arg = sexp_from_stream(io.BytesIO(arg_serialized), to_sexp_f)

            time_parse_input = time.perf_counter()
            cost, result = run_program(
                program, arg, max_cost=max_cost, pre_eval_f=pre_eval_f, strict=args.mempool | args.strict)
            time_done = time.perf_counter()
        if args.cost:
            print("cost = %d" % cost)
        if args.time:
            if args.hex:
                print('read_hex: %f' % (time_read_hex - time_start))
            else:
                print('assemble_from_ir: %f' % (time_assemble - time_start))
                print('to_sexp_f: %f' % (time_parse_input - time_assemble))
            print('run_program: %f' % (time_done - time_parse_input))
        if args.dump:
            blob = as_bin(lambda f: sexp_to_stream(result, f))
            output = blob.hex()
        elif args.quiet:
            output = ''
        else:
            output = binutils.disassemble(result, keywords)
    except EvalError as ex:
        result = to_sexp_f(ex._sexp)
        output = "FAIL: %s %s" % (ex, binutils.disassemble(result, keywords))
        return -1
    except Exception as ex:
        output = str(ex)
        raise
    finally:
        print(output)
        if args.verbose or symbol_table:
            print()
            trace_to_text(log_entries, binutils.disassemble, symbol_table)
        if args.table:
            trace_to_table(log_entries, binutils.disassemble, symbol_table)


def read_ir(args=sys.argv):
    parser = argparse.ArgumentParser(
        description='Read script and tokenize to IR.'
    )
    parser.add_argument(
        "script", help="script in hex or uncompiled text")

    args = parser.parse_args(args=args[1:])

    sexp = reader.read_ir(args.script)
    blob = stream_to_bin(lambda f: sexp_to_stream(sexp, f))
    print(blob.hex())


"""
Copyright 2018 Chik Network Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



from klvm import KEYWORD_TO_ATOM
from klvm_tools.binutils import disassemble
from klvm_tools.NodePath import LEFT, TOP

from .defaults import default_macro_lookup
from .helpers import brun, eval, quote
from .mod import compile_mod

QUOTE_ATOM = KEYWORD_TO_ATOM["q"]
APPLY_ATOM = KEYWORD_TO_ATOM["a"]
CONS_ATOM = KEYWORD_TO_ATOM["c"]

PASS_THROUGH_OPERATORS = set(KEYWORD_TO_ATOM.values())

for _ in "com opt".split():
    PASS_THROUGH_OPERATORS.add(_.encode("utf8"))


def compile_qq(args, macro_lookup, symbol_table, run_program, level=1):
    """
    (qq ATOM) => (q . ATOM)
    (qq (unquote X)) => X
    (qq (a . B)) => (c (qq a) (qq B))
    """

    def com(sexp):
        return do_com_prog(sexp, macro_lookup, symbol_table, run_program)

    sexp = args.first()
    if not sexp.listp() or sexp.nullp():
        # (qq ATOM) => (q . ATOM)
        return sexp.to(quote(sexp))

    if sexp.listp() and not sexp.first().listp():
        op = sexp.first().as_atom()
        if op == b"qq":
            subexp = compile_qq(sexp.rest(), macro_lookup, symbol_table, run_program, level+1)
            return com(sexp.to([CONS_ATOM, op, [CONS_ATOM, subexp, quote(0)]]))
        if op == b"unquote":
            if level == 1:
                # (qq (unquote X)) => X
                return com(sexp.rest().first())
            subexp = compile_qq(sexp.rest(), macro_lookup, symbol_table, run_program, level-1)
            return com(sexp.to([CONS_ATOM, op, [CONS_ATOM, subexp, quote(0)]]))

    # (qq (a . B)) => (c (qq a) (qq B))
    A = com(sexp.to([b"qq", sexp.first()]))
    B = com(sexp.to([b"qq", sexp.rest()]))
    return sexp.to([CONS_ATOM, A, B])


def compile_macros(args, macro_lookup, symbol_table, run_program):
    return args.to(quote(macro_lookup))


def compile_symbols(args, macro_lookup, symbol_table, run_program):
    return args.to(quote(symbol_table))


COMPILE_BINDINGS = {
    b"qq": compile_qq,
    b"macros": compile_macros,
    b"symbols": compile_symbols,
    b"lambda": compile_mod,
    b"mod": compile_mod,
}


# Transform "quote" to "q" everywhere. Note that quote will not be compiled if behind qq.
# Overrides symbol table defns.
def lower_quote(prog, macro_lookup=None, symbol_table=None, run_program=None):
    if prog.nullp():
        return prog

    if prog.listp():
        if prog.first().as_atom() == b"quote":
            # Note: quote should have exactly one arg, so the length of
            # quoted list should be 2: "(quote arg)"
            if not prog.rest().rest().nullp():
                raise SyntaxError("Compilation error while compiling [%s]. quote takes exactly one argument." %
                                      disassemble(prog))
            return prog.to(quote(lower_quote(prog.rest().first())))
        else:
            return prog.to((lower_quote(prog.first()), lower_quote(prog.rest())))
    else:
        return prog

def do_com_prog(prog, macro_lookup, symbol_table, run_program):
    """
    Turn the given program `prog` into a klvm program using
    the macros to do transformation.

    prog is an uncompiled s-expression.

    Return a new expanded s-expression PROG_EXP that is equivalent by rewriting
    based upon the operator, where "equivalent" means

    (a (com (q PROG) (MACROS)) ARGS) == (a (q PROG_EXP) ARGS)
    for all ARGS.

    Also, (opt (com (q PROG) (MACROS))) == (opt (com (q PROG_EXP) (MACROS)))
    """

    # lower "quote" to "q"
    prog = lower_quote(prog, macro_lookup, symbol_table, run_program)

    # quote atoms
    if prog.nullp() or not prog.listp():
        atom = prog.as_atom()
        if atom == b"@":
            return prog.to(TOP.as_path())
        for pair in symbol_table.as_iter():
            symbol, value = pair.first(), pair.rest().first()
            if symbol == atom:
                return prog.to(value)

        return prog.to(quote(prog))

    operator = prog.first()
    if operator.listp():
        # (com ((OP) . RIGHT)) => (a (com (q OP)) 1)
        inner_exp = eval(prog.to([b"com",
            quote(operator), quote(macro_lookup), quote(symbol_table)]), TOP.as_path())
        return prog.to([inner_exp])

    as_atom = operator.as_atom()

    for macro_pair in macro_lookup.as_iter():
        macro_name = macro_pair.first().as_atom()
        if macro_name == as_atom:
            macro_code = macro_pair.rest().first()
            post_prog = brun(macro_code, prog.rest())
            return eval(post_prog.to(
                [b"com", post_prog, quote(macro_lookup), quote(symbol_table)]), TOP.as_short_path())

    if as_atom in COMPILE_BINDINGS:
        f = COMPILE_BINDINGS[as_atom]
        post_prog = f(prog.rest(), macro_lookup, symbol_table, run_program)
        return eval(prog.to(quote(post_prog)), TOP.as_path())

    if operator == QUOTE_ATOM:
        return prog

    compiled_args = [do_com_prog(_, macro_lookup, symbol_table, run_program) for _ in prog.rest().as_iter()]

    r = prog.to([operator] + compiled_args)

    if as_atom in PASS_THROUGH_OPERATORS or as_atom.startswith(b"_"):
        return r

    for (symbol, value) in symbol_table.as_python():
        if symbol == b"*":
            return r
        if symbol == as_atom:
            new_args = eval(
                prog.to([b"opt", [b"com",
                                  quote([b"list"] + list(prog.rest().as_iter())),
                                  quote(macro_lookup),
                                  quote(symbol_table)]]), TOP.as_path())
            r = prog.to([APPLY_ATOM, value, [CONS_ATOM, LEFT.as_path(), new_args]])
            return r

    raise SyntaxError(
        "can't compile %s, unknown operator" %
        disassemble(prog))


def make_do_com(run_program):

    def do_com(sexp):
        prog = sexp.first()
        symbol_table = sexp.null()
        if not sexp.rest().nullp():
            macro_lookup = sexp.rest().first()
            if not sexp.rest().rest().nullp():
                symbol_table = sexp.rest().rest().first()
        else:
            macro_lookup = default_macro_lookup(run_program)
        return 1, do_com_prog(prog, macro_lookup, symbol_table, run_program)

    return do_com



import glob
from numpy import linalg
from numpy import array
from numpy import inner
import os

def to_float_list(l):
    ret = []
    for s in l:
        ret.append(float(s))
    return ret

def linear_regression_no_outliers(ops, runtime):
    num_points = len(ops)

    # we want to remove the 5% worst points
    for counter in range(num_points // 20):
        m, c = linalg.lstsq(ops, runtime)[0]
        # find the point farthest from the line defined by x*m+c
        # and remove it
        worst = -1
        dist = 0
        for i in range(len(ops)):
            f = ops[i][0] * m + c
            d = abs(f - runtime[i])
            if d > dist:
                worst = i
                dist = d

        if worst >= 0:
            del ops[worst]
            del runtime[worst]

    # now, run the regression analysis on the cleaned-up points
    return linalg.lstsq(ops, runtime)[0]

for directory in glob.glob('test-programs/*'):
    gnuplot_filename = os.path.join(directory, 'render-trends.gnuplot')
    gnuplot = open(gnuplot_filename, 'w+')

    gnuplot.write('''set output "%s/timing-trends.png"
set term png size 1400,900 small
set termoption enhanced
set ylabel "run-time (s)"
set xlabel "number of ops"
set xrange [0:3000]
set yrange [0:0.02]
set datafile separator ","
plot ''' % directory)

    print('microseconds per operation')
    color = 0
    for fn in glob.glob('test-programs/%s/results-*.csv' % os.path.split(directory)[1]):
        r = open(fn, 'r')

        ops = []
        runtime = []

        for l in r:
            if l.startswith('#'): continue
            l = l.split(',')
            runtime.append(float(l[3]))
            ops.append([float(l[4]), 1.])

        m, c = linear_regression_no_outliers(ops, runtime)

        name = fn.split('results-')[1].split('.csv')[0]
        print('%30s: %.20f (%f)' % (name, m * 1000000, c * 1000000))

        gnuplot.write('x*%.20f+%.20f title "%s" lc %d, \\\n' % (m, c, name, color))
        color += 1

    gnuplot.write('y=0\n')
    gnuplot.close()
    os.system('gnuplot %s 2>/dev/null' % gnuplot_filename)



import io
import os
import shlex
import sys
import unittest

import importlib_metadata

# If the REPAIR environment variable is set, any tests failing due to
# wrong output will be corrected. Be sure to do a "git diff" to validate that
# you're getting changes you expect.

REPAIR = os.getenv("REPAIR", 0)


def get_test_cases(path):
    PREFIX = os.path.dirname(__file__)
    TESTS_PATH = os.path.join(PREFIX, path)
    paths = []
    for dirpath, dirnames, filenames in os.walk(TESTS_PATH):
        for fn in filenames:
            if fn.endswith(".txt") and fn[0] != ".":
                paths.append(os.path.join(dirpath, fn))
    paths.sort()
    test_cases = []
    for p in paths:
        with open(p) as f:
            # allow "#" comments at the beginning of the file
            cmd_lines = []
            comments = []
            while 1:
                line = f.readline().rstrip()
                if len(line) < 1 or line[0] != "#":
                    if line[-1:] == "\\":
                        cmd_lines.append(line[:-1])
                        continue
                    cmd_lines.append(line)
                    break
                comments.append(line + "\n")
            expected_output = f.read()
            test_name = os.path.relpath(p, PREFIX).replace(".", "_").replace("/", "_")
            test_cases.append((test_name, cmd_lines, expected_output, comments, p))
    return test_cases


class TestCmds(unittest.TestCase):
    def invoke_tool(self, cmd_line):

        # capture io
        stdout_buffer = io.StringIO()
        stderr_buffer = io.StringIO()

        old_stdout = sys.stdout
        old_stderr = sys.stderr

        sys.stdout = stdout_buffer
        sys.stderr = stderr_buffer

        args = shlex.split(cmd_line)
        [entry_point] = importlib_metadata.entry_points(group="console_scripts", name=args[0])
        v = entry_point.load()(args)

        sys.stdout = old_stdout
        sys.stderr = old_stderr

        return v, stdout_buffer.getvalue(), stderr_buffer.getvalue()


def make_f(cmd_lines, expected_output, comments, path):
    def f(self):
        cmd = "".join(cmd_lines)
        for c in cmd.split(";"):
            r, actual_output, actual_stderr = self.invoke_tool(c)
        if actual_output != expected_output:
            print(path)
            print(cmd)
            print(actual_output)
            print(expected_output)
            if REPAIR:
                f = open(path, "w")
                f.write("".join(comments))
                for line in cmd_lines[:-1]:
                    f.write(line)
                    f.write("\\\n")
                f.write(cmd_lines[-1])
                f.write("\n")
                f.write(actual_output)
                f.close()
        self.assertEqual(expected_output, actual_output)

    return f


def inject(*paths):
    for path in paths:
        for idx, (name, i, o, comments, path) in enumerate(get_test_cases(path)):
            name_of_f = "test_%s" % name
            setattr(TestCmds, name_of_f, make_f(i, o, comments, path))


inject("opc")

inject("opd")

inject("stage_1")

inject("stage_2")

inject("klvm_runtime")

inject("cmd")

# inject("v0_0_2")


def main():
    unittest.main()


if __name__ == "__main__":
    main()


"""
Copyright 2018 Chik Network Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



from klvm_tools.binutils import assemble

from stages.stage_0 import run_program

from .pattern_match import match


# CURRY_OBJ_CODE contains compiled code from the output of the following:
# run -i klvm_runtime '(mod (F . args) (include curry.klvm) (curry_args F args))'

# the text below has been hand-optimized to replace `((c (q X) Y))` with `(a (q X) Y)`
# and `(q 0)` with `0`

CURRY_OBJ_CODE = assemble(
    """
(a (q #a 4 (c 2 (c 5 (c 7 0)))) (c (q (c (q . 2) (c (c (q . 1) 5) (c (a 6 (c 2 (c 11 (q 1)))) 0))) #a (i 5 (q 4 (q . 4) (c (c (q . 1) 9) (c (a 6 (c 2 (c 13 (c 11 0)))) 0))) (q . 11)) 1) 1))
    """
)


def curry(program, args):
    """
    ;; A "curry" binds values to a function, making them constant,
    ;; and returning a new function that returns fewer arguments (since the
    ;; arguments are now fixed).
    ;; Example: (defun add2 (V1 V2) (+ V1 V2))  ; add two values
    ;; (curry add2 15) ; this yields a function that accepts ONE argument, and adds 15 to it

    `program`: an SExp
    `args`: an SExp that is a list of constants to be bound to `program`
    """

    args = program.to((program, args))
    r = run_program(CURRY_OBJ_CODE, args)
    return r


UNCURRY_PATTERN_FUNCTION = assemble("(a (q . (: . function)) (: . core))")
UNCURRY_PATTERN_CORE = assemble("(c (q . (: . parm)) (: . core))")


def uncurry(curried_program):
    r = match(UNCURRY_PATTERN_FUNCTION, curried_program)
    if r is None:
        return r

    f = r["function"]
    core = r["core"]

    args = []
    while True:
        r = match(UNCURRY_PATTERN_CORE, core)
        if r is None:
            break
        args.append(r["parm"])
        core = r["core"]

    if core.as_python() == b"\x01":
        return f, f.to(args)
    return None



from ir import reader


def test_reader_1():
    sexp = reader.read_ir("(100 0x0100)")
    print(sexp)

    sexp = reader.read_ir("100")
    print(sexp)

    sexp = reader.read_ir("0x0100")
    print(sexp)

    sexp = reader.read_ir("0x100")
    print(sexp)

    sexp = reader.read_ir('"100"')
    print(sexp)

    sexp = reader.read_ir("foo")
    print(sexp)

    sexp = reader.read_ir('(c (quote 100) (c (quote "foo") (quote ())))')
    print(sexp)

    sexp = reader.read_ir("(c . foo)")
    print(sexp)



# adapted from https://github.com/PyO3/setuptools-rust

from distutils.command.install import install
from distutils.dist import Distribution as DistDistribution
from setuptools.dist import Distribution

try:
    from wheel.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None


def patch_build_ext(build_ext):
    # allow to use 'klvm_extensions' parameter for setup() call
    Distribution.klvm_extensions = ()

    # replace setuptools build_ext
    Distribution.orig_get_command_class = Distribution.get_command_class

    def get_command_class(self, command):
        if command == "build_ext":
            if command not in self.cmdclass:
                self.cmdclass[command] = build_ext

        return self.orig_get_command_class(command)

    Distribution.get_command_class = get_command_class

    # use custom has_ext_modules
    DistDistribution.orig_has_ext_modules = DistDistribution.has_ext_modules

    def has_ext_modules(self):
        return (
            self.ext_modules
            and len(self.ext_modules) > 0
            or self.klvm_extensions
            and len(self.klvm_extensions) > 0
        )

    DistDistribution.has_ext_modules = has_ext_modules

    # this is required because, install directly access distribution's
    # ext_modules attr to check if dist has ext modules
    install.orig_finalize_options = install.finalize_options

    def finalize_options(self):
        ext_modules = self.distribution.ext_modules

        # all ext modules
        mods = []
        if self.distribution.ext_modules:
            mods.extend(self.distribution.ext_modules)
        if self.distribution.klvm_extensions:
            mods.extend(self.distribution.klvm_extensions)

        self.distribution.ext_modules = mods

        self.orig_finalize_options()

        # restore ext_modules
        self.distribution.ext_modules = ext_modules

    install.finalize_options = finalize_options



from klvm import run_program as default_run_program  # noqa
from klvm.operators import OPERATOR_LOOKUP, OperatorDict
from klvm.EvalError import EvalError

from klvm_tools import binutils

def run_program(
    program,
    args,
    operator_lookup=OPERATOR_LOOKUP,
    max_cost=None,
    pre_eval_f=None,
    strict=False,
):
    if strict:
        def fatal_error(op, arguments):
            raise EvalError("unimplemented operator", arguments.to(op))
        operator_lookup = OperatorDict(operator_lookup, unknown_op_handler=fatal_error)

    return default_run_program(
        program,
        args,
        operator_lookup,
        max_cost,
        pre_eval_f=pre_eval_f,
    )


"""
Copyright 2019 Chik Network Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



# klvm_tools setuptools integration

from distutils import log

import os
import pathlib

from ir import reader
from klvm_tools import binutils
from stages import stage_2


def compile_klvm_text(text, search_paths):
    ir_src = reader.read_ir(text)
    assembled_sexp = binutils.assemble_from_ir(ir_src)

    input_sexp = assembled_sexp.to((assembled_sexp, []))
    run_program = stage_2.run_program_for_search_paths(search_paths)
    cost, result = run_program(stage_2.run, input_sexp)
    return result


def compile_klvm(input_path, output_path, search_paths=[]):
    input_path = pathlib.Path(input_path)
    output_path = pathlib.Path(output_path)
    try:
        output_time = output_path.stat().st_mtime
    except FileNotFoundError:
        output_time = None
    if output_time is None or input_path.stat().st_mtime > output_time:
        log.info("klvmcc %s -o %s" % (input_path, output_path))
        with open(input_path) as f:
            text = f.read()
        result = compile_klvm_text(text, search_paths)
        hex = result.as_bin().hex()

        with open(output_path, "w") as f:
            f.write(insert_newlines(hex))
            f.write("\n")
    else:
        log.info("skipping %s, compiled recently" % input_path)

    return output_path


def find_files(path=""):
    r = []
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(".klvm"):
                full_path = pathlib.Path(dirpath, filename)
                target = "%s.hex" % path
                compile_klvm(full_path, target)
                r.append(target)
    return r

def insert_newlines(string, every=80):
    lines = []
    for i in range(0, len(string), every):
        lines.append(string[i:i+every])
    return '\n'.join(lines)



ATOM_MATCH = b"$"
SEXP_MATCH = b":"


def unify_bindings(bindings, new_key, new_value):
    """
    Try to add a new binding to the list, rejecting it if it conflicts
    with an existing binding.
    """
    new_key = bytes(new_key).decode()
    if new_key in bindings:
        if bindings[new_key] != new_value:
            return None
        return bindings
    new_bindings = dict(bindings)
    new_bindings[new_key] = new_value
    return new_bindings


def match(pattern, sexp, known_bindings={}):
    """
    Determine if sexp matches the pattern, with the given known bindings already applied.

    Returns None if no match, or a (possibly empty) dictionary of bindings if there is a match

    Patterns look like this:

    ($ . $) matches the literal "$", no bindings (mostly useless)
    (: . :) matches the literal ":", no bindings (mostly useless)

    ($ . A) matches B iff B is an atom; and A is bound to B
    (: . A) matches B always; and A is bound to B

    (A . B) matches (C . D) iff A matches C and B matches D
          and bindings are the unification (as long as unification is possible)
    """

    if not pattern.listp():
        if sexp.listp():
            return None
        return known_bindings if pattern.as_atom() == sexp.as_atom() else None

    left = pattern.first()
    right = pattern.rest()
    atom = sexp.as_atom()

    if left == ATOM_MATCH:
        if sexp.listp():
            return None
        if right == ATOM_MATCH:
            if atom == ATOM_MATCH:
                return {}
            return None
        return unify_bindings(known_bindings, right.as_atom(), sexp)

    if left == SEXP_MATCH:
        if right == SEXP_MATCH:
            if atom == SEXP_MATCH:
                return {}
            return None
        return unify_bindings(known_bindings, right.as_atom(), sexp)

    if not sexp.listp():
        return None

    new_bindings = match(left, sexp.first(), known_bindings)
    if new_bindings is None:
        return new_bindings
    return match(right, sexp.rest(), new_bindings)



import os
import random
import sys

def make_value(length):
    ret = '0x'
    for i in range(length):
        ret += '%02x' % ((random.getrandbits(7) << 1) + 1)
    return ret

def make_lookup(depth):
    path = 1
    tree = '42';
    while depth > 0:
        path <<= 1
        leg = random.getrandbits(1)
        if leg == 0:
            tree = '(' + tree + ' . 0x1337)'
        else:
            tree = '(0x1337 . ' + tree + ')'
        path |= leg
        depth -= 1
    return '%d' % path, tree

def generate_args(n, name, value_size, filename):
    ret = '(' + name
    for i in range(n):
        ret += ' (q . ' + make_value(value_size) + ')'
    ret += ')'
    return '%s_args-%d-%d' % (filename, value_size, n), ret, '()'

def generate_nested(n, name, value_size, filename, arity=2):
    ret = ''
    for i in range(n):
        ret += '(%s ' % name
        for i in range(arity - 1):
            ret += '(q . ' + make_value(value_size) + ') '
    ret += '(q . ' + make_value(value_size) + ')'
    for i in range(n):
        ret += ')'
    return '%s_nest-%d-%d' % (filename, value_size, n), ret, '()'

def generate_nested_1(n, name, value_size, filename, arity=2):
    ret = ''
    for i in range(n):
        ret += '(%s ' % name
        for i in range(arity - 1):
            ret += '(q . 1) '
    ret += '(q . ' + make_value(value_size) + ')'
    for i in range(n):
        ret += ')'
    return '%s_nest1-%d-%d' % (filename, value_size, n), ret, '()'

def size_of_value(val):
    if val.startswith('0x'):
        return (len(val) - 2) / 2
    if val.startswith('"') and val.endswith('"'):
        return len(val) - 2
    print("don't know how to interpret value: %s" % val)
    sys.exit(1)

def generate_args_value(n, name, value, filename):
    ret = '(' + name
    for i in range(n):
        ret += ' (q . ' + value + ')'
    ret += ')'
    return '%s_args-%d-%d' % (filename, size_of_value(value), n), ret, '()'

def generate_nested_value(n, name, value, filename, arity=2):
    ret = ''
    for i in range(n):
        ret += '(%s ' % name
        for i in range(arity - 1):
            ret += '(q . ' + value + ') '
    ret += '(q . ' + value + ')'
    for i in range(n):
        ret += ')'
    return '%s_nest-%d-%d' % (filename, size_of_value(value), n), ret, '()'

# use a different value for right hand and left hand. e.g. shift has limits on
# how large the right hand side can be
def generate_nested_2values(n, name, value_sizes, filename, arity=2):
    ret = ''
    for i in range(n):
        ret += '(%s ' % name
    ret += '(q . ' + make_value(value_sizes[0]) + ')'
    for i in range(n):
        for i in range(arity - 1):
            ret += ' (q . ' + make_value(value_sizes[1]) + ')'
        ret += ')'
    return '%s_nest-%d-%d' % (filename, value_sizes[0], n), ret, '()'

def generate_lookup(n):
    path, tree = make_lookup(n)
    return 'lookup-2-%d' % n, path, tree

def generate_lookup_op_list(n):
    path, tree = make_lookup(1)
    ret = ''
    for i in range(n):
        ret += '(c ' + path + ' '
    ret += '()'
    for i in range(n):
        ret += ')'

    return 'lookup_2-2-%d' % n, ret, tree

def generate_op_list(n, name, value_size, filename, arity=2):
    ret = ''
    for i in range(n):
        ret += '(c (%s' % name
        for i in range(arity):
            ret += ' (q . ' + make_value(value_size) + ')'
        ret += ') '
    ret += '()'
    for i in range(n):
        ret += ')'
    return '%s-%d-%d' % (filename, value_size, n), ret, '()'

def generate_list(n, name, filename):
    ret = ''
    for i in range(n):
        ret += '(c (%s (q . (1 2 3))) ' % name
    ret += '()'
    for i in range(n):
        ret += ')'
    return '%s-1-%d' % (filename, n), ret, '()'

def generate_list_empty(n):
    ret = ''
    for i in range(n):
        ret += '(c (q . (1 2 3)) '
    ret += '()'
    for i in range(n):
        ret += ')'
    return 'first_empty-1-%d' % n, ret, '()'

def generate_if(n):
    ret = ''
    # alternate between true and false
    conditions = ['()', '(q . 1)']
    for i in range(n):
        ret += '(c (i %s (q . 1) (q . 2)) ' % conditions[i % 2]
    ret += '()'
    for i in range(n):
        ret += ')'
    return 'if-1-%d' % n, ret, '()'

def gen_apply(n, name):
    folder = 'test-programs/%s' % name
    try: os.mkdir(folder)
    except: pass
    with open(folder + '/%s-%d.klvm' % (name, n), 'w+') as f:
        for i in range(n):
            f.write('(a (q . (lognot ');

        f.write('(q . 1)');

        for i in range(n):
            f.write(')) ())');

    with open(folder + '/%s-%d.env' % (name, n), 'w+') as f:
        f.write('()')

def get_range(name):
    if name.split('-')[0].endswith('_empty'): return 3000,40,[1]
    if name.startswith('mul_nest1'): return 3000,300,[1, 25, 50, 100, 200, 400, 600, 800, 1000]
    if name.startswith('mul'): return 3000,50,[1, 25, 50, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200, 1300, 1400]
    if name.startswith('cons-'): return 3000,40,[1]
    if name.startswith('lookup'): return 3000,40,[2]
    if name.startswith('point_add'): return 300,4,[48]
    if name.startswith('listp'): return 3000,40,[1]
    if name.startswith('first'): return 3000,40,[1]
    if name.startswith('rest'): return 3000,40,[1]
    if name.startswith('if-'): return 3000,40,[1]
    else: return 3000,40,[1, 128, 1024]

def print_files(fun):
    name = fun(0, 1)[0]
    end, step, vsizes = get_range(name)
    folder = 'test-programs/' + name.split('-')[0].split('_')[0]
    try: os.mkdir(folder)
    except: pass
    for value_size in vsizes:
        for i in range(2, end, step):
            name, prg, env = fun(i, value_size)
            open(folder + '/' + name + '.klvm', 'w+').write(prg)
            open(folder + '/' + name + '.env', 'w+').write(env)

try: os.mkdir('test-programs')
except: pass

print_files(lambda n, vs: generate_op_list(n, 'concat', vs, 'concat'))
print_files(lambda n, vs: generate_args(n, 'concat', vs, 'concat'))

print_files(lambda n, vs: generate_op_list(n, 'divmod', vs, 'divmod'))
print_files(lambda n, vs: generate_op_list(n, '/', vs, 'div'))

print_files(lambda n, vs: generate_args(n, '+', vs, 'plus'))
print_files(lambda n, vs: generate_op_list(n, '+', vs, 'plus'))
print_files(lambda n, vs: generate_op_list(n, '+', vs, 'plus_empty', arity=0))

print_files(lambda n, vs: generate_args(n, '-', vs, 'minus'))
print_files(lambda n, vs: generate_op_list(n, '-', vs, 'minus'))
print_files(lambda n, vs: generate_op_list(n, '-', vs, 'minus_empty', arity=0))

print_files(lambda n, vs: generate_args(n, 'logand', vs, 'logand'))
print_files(lambda n, vs: generate_op_list(n, 'logand', vs, 'logand'))
print_files(lambda n, vs: generate_op_list(n, 'logand', vs, 'logand_empty', arity=0))

print_files(lambda n, vs: generate_args(n, 'logior', vs, 'logior'))
print_files(lambda n, vs: generate_op_list(n, 'logior', vs, 'logior'))
print_files(lambda n, vs: generate_op_list(n, 'logior', vs, 'logior_empty', arity=0))

print_files(lambda n, vs: generate_args(n, 'logxor', vs, 'logxor'))
print_files(lambda n, vs: generate_op_list(n, 'logxor', vs, 'logxor'))
print_files(lambda n, vs: generate_op_list(n, 'logxor', vs, 'logxor_empty', arity=0))

print_files(lambda n, vs: generate_nested(n, 'lognot', vs, 'lognot', arity=1))
print_files(lambda n, vs: generate_nested(n, 'not', vs, 'not', arity=1))

print_files(lambda n, vs: generate_nested(n, 'any', vs, 'any'))
print_files(lambda n, vs: generate_args(n, 'any', vs, 'any'))

print_files(lambda n, vs: generate_nested(n, 'all', vs, 'all'))
print_files(lambda n, vs: generate_args(n, 'all', vs, 'all'))

print_files(lambda n, vs: generate_nested_2values(n, 'lsh', [vs, 1], 'lsh'))
print_files(lambda n, vs: generate_nested_2values(n, 'ash', [vs, 1], 'ash'))

print_files(lambda n, vs: generate_op_list(n, '=', vs, 'eq'))
print_files(lambda n, vs: generate_op_list(n, '>', vs, 'gr'))
print_files(lambda n, vs: generate_op_list(n, '>s', vs, 'grs'))

print_files(lambda n, vs: generate_nested(n, 'c', vs, 'cons'))

point_val = '0xb3b8ac537f4fd6bde9b26221d49b54b17a506be147347dae5d081c0a6572b611d8484e338f3432971a9823976c6a232b'
print_files(lambda n, vs: generate_nested_value(n, 'point_add', point_val, 'point_add'))
print_files(lambda n, vs: generate_args_value(n, 'point_add', point_val, 'point_add'))

print_files(lambda n, vs: generate_op_list(n, 'sha256', vs, 'sha', arity=1))
print_files(lambda n, vs: generate_op_list(n, 'sha256', vs, 'sha_empty', arity=0))
print_files(lambda n, vs: generate_args(n, 'sha256', vs, 'sha'))

print_files(lambda n, vs: generate_op_list(n, 'pubkey_for_exp', vs, 'pubkey', arity=1))

print_files(lambda n, vs: generate_lookup(n))
print_files(lambda n, vs: generate_lookup_op_list(n))

print_files(lambda n, vs: generate_op_list(n, 'strlen', vs, 'strlen', arity=1))

print_files(lambda n, vs: generate_op_list(n, '*', vs, 'mul'))
print_files(lambda n, vs: generate_nested_1(n, '*', vs, 'mul'))
print_files(lambda n, vs: generate_op_list(n, '*', vs, 'mul_empty', arity=0))

print_files(lambda n, vs: generate_op_list(n, 'l', vs, 'listp', arity=1))

print_files(lambda n, vs: generate_list(n, 'f', 'first'))
print_files(lambda n, vs: generate_list_empty(n))

print_files(lambda n, vs: generate_list(n, 'r', 'rest'))

print_files(lambda n, vs: generate_if(n))
gen_apply(1000, 'apply')




#!/usr/bin/env python

from setuptools import setup

with open("README.md", "rt") as fh:
    long_description = fh.read()

dependencies = [
    "klvm>=0.9.2",
    "klvm_tools_rs>=0.1.37",
    "importlib_metadata",
    "setuptools",
]

dev_dependencies = [
    "pytest",
]

setup(
    name="klvm_tools",
    packages=[
        "ir",
        "klvm_tools",
        "klvm_tools.setuptools",
        "stages",
        "stages.stage_2",
    ],
    author="Chik Network, Inc.",
    entry_points={
        "console_scripts": [
            "read_ir = klvm_tools.cmds:read_ir",
            "opc = klvm_tools.cmds:opc",
            "opd = klvm_tools.cmds:opd",
            "run = klvm_tools.cmds:run",
            "brun = klvm_tools.cmds:brun",
        ],
    },
    package_data={
        "": ["py.typed"],
    },
    author_email="hello@chiknetwork.com",
    install_requires=dependencies,
    url="https://github.com/Chik-Network",
    license="Apache-2.0",
    description="KLVM compiler.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Security :: Cryptography",
    ],
    extras_require=dict(
        dev=dev_dependencies,
    ),
    project_urls={
        "Bug Reports": "https://github.com/Chik-Network/klvm_tools",
        "Source": "https://github.com/Chik-Network/klvm_tools",
    },
)



"""
These tests check that the `klvmc` utility methods
continue to work with the `include` keyword, and produce
the expected output. It's not intended to be a complete
test of the compiler, just the `klvmc` api.
"""

from tempfile import TemporaryDirectory

from klvm_tools import klvmc


INCLUDE_CODE = "((defconstant FOO 6001))"
MAIN_CODE = """(mod (VALUE) (include "include.klvm") (+ VALUE FOO))"""
EXPECTED_HEX_OUTPUT = "ff10ff02ffff0182177180"

# `EXPECTED_HEX_OUTPUT` disassembles to "(+ 2 (q . 6001))"


def test_compile_klvm_text():
    with TemporaryDirectory() as include_dir:
        include_path = f"{include_dir}/include.klvm"
        with open(include_path, "w") as f:
            f.write(INCLUDE_CODE)
        output = klvmc.compile_klvm_text(MAIN_CODE, search_paths=[include_dir])
        assert repr(output) == f"SExp({EXPECTED_HEX_OUTPUT})"


def test_compile_klvm():
    with TemporaryDirectory() as include_dir:
        with TemporaryDirectory() as source_dir:
            with open(f"{include_dir}/include.klvm", "w") as f:
                f.write(INCLUDE_CODE)
            main_path = f"{source_dir}/main.klvm"
            main_output = f"{source_dir}/main.hex"
            with open(main_path, "w") as f:
                f.write(MAIN_CODE)
            output = klvmc.compile_klvm(
                main_path, main_output, search_paths=[include_dir]
            )
            t = open(output).read()
            assert t == f"{EXPECTED_HEX_OUTPUT}\n"



import glob
import subprocess
import sys
import os

open_files = {}

dry_run = False

def get_file(folder, name):
    full_path = os.path.join(folder, 'results-%s.csv' % name)
    if full_path in open_files: return open_files[full_path]

    if dry_run: f = 1
    else: f = open(full_path, 'a')
    open_files[full_path] = f

    if dry_run: return f

    f.write('#cost,assemble_from_ir,to_sexp_f,run_program,multiplier\n')
    return f

counter = 0

force_run = '--force' in sys.argv

for directory in glob.glob('test-programs/*'):
    existing_results = []
    for r in glob.glob(directory + '/*.csv'):
        existing_results.append(os.path.split(r)[1].split('-')[1])

    for fn in glob.glob('test-programs/%s/*.klvm' % os.path.split(directory)[1]):

        # if we have a csv file for this run already, skip running it again
        dry_run = not force_run and os.path.split(fn)[1].split('-')[0] in existing_results

        if not dry_run:
            print('%04d: %s' % (counter, fn))
        counter += 1
        # the filename is expected to be in the form:
        # name "-" value_size "-" num_calls
        if not dry_run:
            env = open(fn[:-4] + 'env').read()
            output = subprocess.check_output(['brun', '--backend=rust', '-c', '--quiet', '--time', fn, env])
            output = output.decode('ascii').split('\n', 5)[:-1]

            counters = {}
            for o in output:
                try:
                    if ':' in o:
                        key, value = o.split(':')
                        counters[key.strip()] = value.strip()
                    elif '=' in o:
                        key, value = o.split('=')
                        counters[key.strip()] = value.strip()
                except BaseException as e:
                    print(e)
                    print('ERROR parsing: %s' % o)
            print(counters)

        folder, fn = os.path.split(fn)
        name_components = fn.split('-')
        f = get_file(folder, '-'.join(name_components[0:-1]))
        if not dry_run:
            line = counters['cost'] + ',' + \
                counters['assemble_from_ir'] + ',' + \
                counters['to_sexp_f'] + ',' + \
                counters['run_program'] + ',' + \
                name_components[-1].split('.')[0] + '\n'
            f.write(line)

    gnuplot_filename = '%s/render-timings.gnuplot' % folder
    gnuplot = open(gnuplot_filename, 'w+')

    gnuplot.write('''set output "%s/timings.png"
set datafile separator ","
set term png size 1400,900 small
set termoption enhanced
set ylabel "run-time (s)"
set xlabel "number of ops"
set xrange [0:*]
set yrange [0:0.3]
''' % directory)

    color = 0
    gnuplot.write('plot ')
    count = len(open_files)
    for n,v in open_files.items():
        cont = ', \\'
        if color + 1 == count: cont = ''
        name = n.split('results-')[1].split('.csv')[0]
        gnuplot.write('"%s" using 5:4 with points lc %d title "%s"%s\n' % (n, color, name, cont))
        color += 1
        if not isinstance(v, int): v.close()

    gnuplot.close()
    open_files = {}
    os.system('gnuplot %s' % gnuplot_filename)



from klvm import KEYWORD_TO_ATOM
from klvm_tools.NodePath import TOP


QUOTE_ATOM = KEYWORD_TO_ATOM["q"]
APPLY_ATOM = KEYWORD_TO_ATOM["a"]


def quote(sexp):
    """quoted list as a python list, not as an sexp"""
    return (QUOTE_ATOM, sexp)


def eval(prog, args):
    return prog.to([APPLY_ATOM, prog, args])


def run(prog, macro_lookup):
    """
    PROG => (e (com (q . PROG) (mac)) ARGS)

    The result can be evaluated with the stage_com eval
    function.
    """
    args = TOP.as_path()
    mac = quote(macro_lookup)
    return eval(prog.to([b"com", prog, mac]), args)


def brun(prog, args):
    return eval(prog.to(quote(prog)), quote(args))



from distutils.cmd import Command
from distutils import log

from setuptools.dist import Distribution

from klvm_tools.klvmc import compile_klvm


Distribution.klvm_extensions = ()


class build_klvm(Command):
    """ Command for building klvm """

    description = "build klvm extensions"

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        file_list = self.distribution.klvm_extensions
        for _ in file_list:
            log.info("build_klvm on %s" % _)
            target = "%s.hex" % _
            compile_klvm(_, target)



from distutils import log
from setuptools.command.build_ext import build_ext as _build_ext


class build_ext(_build_ext):

    def __init__(self, *args):
        _build_ext.__init__(self, *args)

    def has_klvm_extensions(self):
        return (
            self.distribution.klvm_extensions
            and len(self.distribution.klvm_extensions) > 0
        )

    def check_extensions_list(self, extensions):
        if extensions:
            _build_ext.check_extensions_list(self, extensions)

    def run(self):
        """Run build_klvm sub command """
        if self.has_klvm_extensions():
            log.info("running build_klvm")
            build_klvm = self.get_finalized_command("build_klvm")
            build_klvm.inplace = self.inplace
            build_klvm.run()

        _build_ext.run(self)



from klvm.casts import int_from_bytes

from klvm_tools.NodePath import NodePath, TOP, LEFT, RIGHT

LEFT_RIGHT_LEFT = LEFT + RIGHT + LEFT


def reset(n):
    path_blob = n.as_short_path()
    index = int_from_bytes(path_blob)
    n = NodePath(index)
    return n


def cmp_to_bits(n, bits):
    n_as_int = int(f"0x{n.as_short_path().hex()}", 16)
    bits_as_int = int(f"0b{bits}", 2)
    assert n_as_int == bits_as_int


def test_node_path():
    n = TOP
    assert n.as_short_path().hex() == "01"
    n = reset(n)
    n += LEFT
    assert n.as_short_path().hex() == "02"
    n = reset(n)
    n += RIGHT
    assert n.as_short_path().hex() == "06"
    n = reset(n)
    n += RIGHT
    assert n.as_short_path().hex() == "0e"
    n = reset(n)
    n += LEFT
    assert n.as_short_path().hex() == "16"
    n = reset(n)
    n += LEFT
    assert n.as_short_path().hex() == "26"
    n = reset(n)
    n += LEFT
    assert n.as_short_path().hex() == "46"
    n = reset(n)
    n += RIGHT
    assert n.as_short_path().hex() == "c6"
    n = reset(n)
    n += LEFT
    cmp_to_bits(n, "101000110")
    n = reset(n)
    n += LEFT_RIGHT_LEFT
    cmp_to_bits(n, "101001000110")
    n = reset(n)
    n += LEFT_RIGHT_LEFT
    cmp_to_bits(n, "101001001000110")
    n = reset(n)
    n += LEFT_RIGHT_LEFT
    cmp_to_bits(n, "101001001001000110")
    n = reset(n)
    cmp_to_bits(n, "101001001001000110")


def test_revive_index():
    for idx in range(2048):
        n = NodePath(idx)
        n1 = reset(n)
        assert n.as_short_path() == n1.as_short_path()
